"""
@file
@brief Helpers to investigate a tree structure.
"""
import numpy
from sklearn.tree._tree import TREE_LEAF  # pylint: disable=E0611


def _get_tree(obj):
    """
    Returns the tree object.
    """
    if hasattr(obj, "children_left"):
        return obj
    if hasattr(obj, "tree_"):
        return obj.tree_
    raise AttributeError("obj is no tree: {}".format(type(obj)))


def tree_leave_index(model):
    """
    Returns the indices of every leave in a tree.

    @param      model       something which has a member ``tree_``
    @return                 leave indices
    """
    tree = _get_tree(model)
    res = []
    for i in range(tree.node_count):
        if tree.children_left[i] == TREE_LEAF:
            res.append(i)
    return res


def tree_find_path_to_root(tree, i, parents=None):
    """
    Lists nodes involved into the path to find node *i*.

    @param      tree        tree
    @param      i           node index (``tree.nodes[i]``)
    @param      parents     precomputed parents (None -> calls @see fn tree_node_range)
    @return                 one array of size *(D, 2)* where *D* is the number of dimensions
    """
    tree = _get_tree(tree)
    path_i = [i]
    current_i = i
    while current_i in parents:
        current_i = parents[current_i]
        if current_i < 0:
            current_i = - current_i
        path_i.append(current_i)
    return list(reversed(path_i))


def tree_find_common_node(tree, i, j, parents=None):
    """
    Finds the common node to nodes *i* and *j*.

    @param      tree        tree
    @param      i           node index (``tree.nodes[i]``)
    @param      j           node index (``tree.nodes[j]``)
    @param      parents     precomputed parents (None -> calls @see fn tree_node_range)
    @return                 common root, remaining path to *i*, remaining path to *j*
    """
    tree = _get_tree(tree)
    if parents is None:
        parents = tree_node_parents(tree)
    path_i = tree_find_path_to_root(tree, i, parents)
    path_j = tree_find_path_to_root(tree, j, parents)
    for pos, (a, b) in enumerate(zip(path_i, path_j)):
        if a != b:
            return a, path_i[pos:], path_j[pos:]
    pi = parents.get(i, None)
    pj = parents.get(j, None)
    pos = min(len(path_i), len(path_j))
    if pi is not None and pi == j:
        return j, path_i[pos:], path_j[pos:]
    if pj is not None and pj == i:
        return i, path_i[pos:], path_j[pos:]
    raise RuntimeError(  # pragma: no cover
        "Paths are equal, i={} and j={} must be differet.".format(i, j))


def tree_node_parents(tree):
    """
    Returns a dictionary ``{node_id: parent_id}``.

    @param      tree        tree
    @return                 parents
    """
    tree = _get_tree(tree)
    parents = {}
    for i in range(tree.node_count):
        if tree.children_left[i] == TREE_LEAF:
            continue
        parents[tree.children_left[i]] = i
        parents[tree.children_right[i]] = -i
    return parents


def tree_node_range(tree, i, parents=None):
    """
    Determines the ranges for a node all dimensions.
    ``nan`` means infinity.

    @param      tree        tree
    @param      i           node index (``tree.nodes[i]``)
    @param      parents     precomputed parents (None -> calls @see fn tree_node_range)
    @return                 one array of size *(D, 2)* where *D* is the number of dimensions

    The following example shows what the function returns
    in case of simple grid in two dimensions.

    .. runpython::
        :showcode:

        import numpy
        from sklearn.tree import DecisionTreeClassifier
        from mlinsights.mltree import tree_leave_index, tree_node_range

        X = numpy.array([[0, 0], [0, 1], [0, 2],
                         [1, 0], [1, 1], [1, 2],
                         [2, 0], [2, 1], [2, 2]])
        y = list(range(X.shape[0]))
        clr = DecisionTreeClassifier(max_depth=4)
        clr.fit(X, y)

        leaves = tree_leave_index(clr)
        ra = tree_node_range(clr, leaves[0])

        print(ra)
    """
    tree = _get_tree(tree)
    if parents is None:
        parents = tree_node_parents(tree)
    path = tree_find_path_to_root(tree, i, parents)
    mx = max([tree.feature[p] for p in path])
    res = numpy.full((mx + 1, 2), numpy.nan)
    for ind, p in enumerate(path):
        if p == i:
            break
        fn = tree.feature[p]
        lr = tree.children_left[p] == path[ind + 1]
        th = tree.threshold[p]
        if lr:
            res[fn, 1] = min(res[fn, 1], th) if not numpy.isnan(
                res[fn, 1]) else th
        else:
            res[fn, 0] = max(res[fn, 0], th) if not numpy.isnan(
                res[fn, 0]) else th
    return res


def predict_leaves(model, X):
    """
    Returns the leave every observations of *X*
    falls into.

    @param      model       a decision tree
    @param      X           observations
    @return                 array of leaves
    """
    if hasattr(model, 'get_leaves_index'):
        leaves_index = model.get_leaves_index()
    else:
        leaves_index = [i for i in range(len(model.tree_.children_left))
                        if model.tree_.children_left[i] == TREE_LEAF]
    leaves = model.decision_path(X)
    leaves = leaves[:, leaves_index]
    mat = numpy.argmax(leaves, 1)
    res = numpy.asarray(mat).ravel()
    res = numpy.array([leaves_index[r] for r in res])
    return res


def tree_leave_neighbors(model):
    """
    The function determines which leaves are neighbors.
    The method uses some memory as it creates creates a
    grid of the feature spaces, each split multiplies the
    number of cells by two.

    @param      model       a :epkg:`sklearn:tree:DecisionTreeRegressor`,
                            a :epkg:`sklearn:tree:DecisionTreeClassifier`,
                            a model which has a member ``tree_``
    @return                 a dictionary ``{(i, j): (dimension, x1, x2)}``,
                            *i, j* are node indices, if :math:`X_d * sign < th  * sign`,
                            the observations goes to node *i*, *j* otherwise,
                            *i < j*. The border is somewhere in the segment ``[x1, x2]``.

    The following example shows what the function returns
    in case of simple grid in two dimensions.

    .. runpython::
        :showcode:

        import numpy
        from sklearn.tree import DecisionTreeClassifier
        from mlinsights.mltree import tree_leave_neighbors

        X = numpy.array([[0, 0], [0, 1], [0, 2],
                         [1, 0], [1, 1], [1, 2],
                         [2, 0], [2, 1], [2, 2]])
        y = list(range(X.shape[0]))
        clr = DecisionTreeClassifier(max_depth=4)
        clr.fit(X, y)

        nei = tree_leave_neighbors(clr)

        import pprint
        pprint.pprint(nei)
    """
    tree = _get_tree(model)

    # creates the coordinates of the grid

    features = {}
    for i in range(tree.node_count):
        fe = tree.feature[i]
        if fe < 0:
            # leave
            continue
        th = tree.threshold[i]
        if fe not in features:
            features[fe] = []
        features[fe].append(th)
    for fe in features:
        features[fe] = list(sorted(set(features[fe])))
    for fe, v in features.items():
        if len(v) == 1:
            d = abs(v[0]) / 10
            if d == v[0]:
                d = 1
            v.insert(0, v[0] - d)
            v.append(v[-1] + d)
        else:
            diff = [v[i + 1] - v[i] for i in range(len(v) - 1)]
            mdiff = min(diff)
            v.append(v[-1] + mdiff)
            v.insert(0, v[0] - mdiff)

    # predictions

    keys = list(sorted(features))
    pos = [0 for k in keys]
    shape = [len(features[k]) - 1 for k in keys]
    cells = numpy.full(shape, 0, numpy.int32)
    while pos[0] < len(features[keys[0]]) - 1:
        # evaluate
        xy = numpy.zeros((1, model.n_features_))
        for p, k in zip(pos, keys):
            xy[0, k] = (features[k][p] + features[k][p + 1]) / 2
        leave = predict_leaves(model, xy)
        cells[tuple(pos)] = leave[0]

        # next
        ind = len(pos) - 1
        pos[ind] += 1
        while ind > 0 and pos[ind] >= len(features[keys[ind]]) - 1:
            pos[ind] = 0
            ind -= 1
            pos[ind] += 1

    # neighbors

    neighbors = {}
    pos = [0 for k in keys]
    while pos[0] <= len(features[keys[0]]) - 1:
        # neighbors
        try:
            cl = cells[tuple(pos)]
        except IndexError:
            # outside the cube
            cl = None
        if cl is not None:
            for k in range(len(pos)):  # pylint: disable=C0200
                pos[k] += 1
                try:
                    cl2 = cells[tuple(pos)]
                except IndexError:
                    # outside the cube
                    pos[k] -= 1
                    continue
                if cl != cl2:
                    edge = (cl, cl2) if cl < cl2 else (cl2, cl)
                    if edge not in neighbors:
                        neighbors[edge] = []
                    xy = numpy.zeros((model.n_features_))
                    for p, f in zip(pos, keys):
                        xy[f] = (features[f][p] + features[f][p + 1]) / 2
                    x2 = tuple(xy)
                    pos[k] -= 1
                    p = pos[k]
                    key = keys[k]
                    xy[key] = (features[key][p] + features[key][p + 1]) / 2
                    x1 = tuple(xy)
                    neighbors[edge].append((key, x1, x2))
                else:
                    pos[k] -= 1

        # next

        ind = len(pos) - 1
        pos[ind] += 1
        while ind > 0 and pos[ind] >= len(features[keys[ind]]) - 1:
            pos[ind] = 0
            ind -= 1
            pos[ind] += 1

    return neighbors



"""
@file
@brief Dummy auto-regressor which takes past values as predictions.
"""
import numpy
from .base import BaseTimeSeries, TimeSeriesRegressorMixin
from .utils import check_ts_X_y


class DummyTimeSeriesRegressor(BaseTimeSeries, TimeSeriesRegressorMixin):
    """
    Dummy regressor for time series. Use past values as prediction.
    """

    def __init__(self, estimator="dummy", past=1, delay1=1, delay2=2,
                 use_all_past=False, preprocessing=None):
        """
        @param      estimator       estimator to use for regression,
                                    :epkg:`sklearn:linear_model:LinearRegression`
                                    implements a linear auto-regressor,
                                    ``'dummy'`` use past value as predictions
        @param      past            values to use to predict
        @param      delay1          the model computes the first prediction for
                                    *time=t + delay1*
        @param      delay2          the model computes the last prediction for
                                    *time=t + delay2* excluded
        @param      use_all_past    use all past features, not only the timeseries
        @param      preprocessing   preprocessing to apply before predicting,
                                    only the timeseries itselves, it can be
                                    a difference, it must be of type
                                    @see cl BaseReciprocalTimeSeriesTransformer
        """
        TimeSeriesRegressorMixin.__init__(self)
        BaseTimeSeries.__init__(self, past=past, delay1=delay1, delay2=delay2,
                                use_all_past=use_all_past, preprocessing=preprocessing)

    def fit(self, X, y, sample_weight=None):
        """
        Trains the model.

        :param X: output of
            X may be empty (None)
        :param y: timeseries (one single vector), array [n_obs]
        :param sample_weight: weights None or array [n_obs]
        :return: self
        """
        X, y, _ = self._base_fit_predict(X, y, sample_weight)
        check_ts_X_y(self, X, y)
        return self

    def predict(self, X, y):
        """
        Returns the prediction
        """
        X, y, _ = self._base_fit_predict(X, y, None)
        check_ts_X_y(self, X, y)
        nbrow = X.shape[0]
        X, y = self._applies_preprocessing(X, y, None)[:2]
        nb = self.delay2 - self.delay1
        pred = numpy.empty((nbrow, nb), dtype=X.dtype)
        first = nbrow - X.shape[0]
        pred[:first] = numpy.nan
        for i in range(0, nb):
            pred[first:, i] = X[:, -1]
        return pred



"""
@file
@brief Implements a piecewise linear regression.
"""
import numpy
import numpy.random
import pandas
from sklearn.base import BaseEstimator, RegressorMixin, ClassifierMixin, clone
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.preprocessing import KBinsDiscretizer
from sklearn.utils._joblib import Parallel, delayed
from sklearn.utils.fixes import _joblib_parallel_args
try:
    from tqdm import tqdm
except ImportError:
    pass


def _fit_piecewise_estimator(i, model, X, y, sample_weight, association, nb_classes, random_state):
    ind = association == i
    if not numpy.any(ind):
        # No training example for this bucket.
        return None
    Xi = X[ind, :]
    yi = y[ind]
    sw = sample_weight[ind] if sample_weight is not None else None

    if nb_classes is not None and len(set(yi)) != nb_classes:
        # Issues a classifiers requires to have at least one example
        # of each class.
        if random_state is None:
            random_state = numpy.random.RandomState()  # pylint: disable=E1101
        addition = numpy.arange(len(ind))
        random_state.shuffle(addition)
        found = set(yi)
        allcl = set(y)
        res = []
        while len(found) < len(allcl):
            for ki in addition:
                if y[ki] not in found:
                    res.append(ki)
                    found.add(y[ki])
        ind = ind.copy()
        for ki in res:
            ind[ki] = True

        Xi = X[ind, :]
        yi = y[ind]
        sw = sample_weight[ind] if sample_weight is not None else None

    return model.fit(Xi, yi, sample_weight=sw)


def _predict_piecewise_estimator(i, est, X, association):
    ind = association == i
    if not numpy.any(ind):
        return None, None
    return ind, est.predict(X[ind, :])


def _predict_proba_piecewise_estimator(i, est, X, association):
    ind = association == i
    if not numpy.any(ind):
        return None, None
    return ind, est.predict_proba(X[ind, :])


def _decision_function_piecewise_estimator(i, est, X, association):
    ind = association == i
    if not numpy.any(ind):
        return None, None
    return ind, est.decision_function(X[ind, :])


class PiecewiseEstimator(BaseEstimator):
    """
    Uses a :epkg:`decision tree` to split the space of features
    into buckets and trains a linear regression on each of them.
    The second estimator can be a :epkg:`sklearn:linear_model:LinearRegression`
    for a regression or :epkg:`sklearn:linear_model:LogisticRegression`
    for a classifier. It can also be :epkg:`sklearn:dummy:DummyRegressor`
    :epkg:`sklearn:dummy:DummyClassifier` to just get the average on each bucket.
    When the buckets are defined by a decision tree and the
    estimator is linear, @see cl PiecewiseTreeRegressor optimizes
    the buckets based on the results of a linear regression.
    The accuracy is usually better.
    """

    def __init__(self, binner=None, estimator=None, n_jobs=None, verbose=False):
        """
        @param      binner              transformer or predictor which creates the buckets
        @param      estimator           predictor trained on every bucket
        @param      n_jobs              number of parallel jobs (for training and predicting)
        @param      verbose             boolean or use ``'tqdm'`` to use :epkg:`tqdm`
                                        to fit the estimators

        *binner* must be filled or must be:

        - ``'bins'``: the model :epkg:`sklearn:preprocessing:KBinsDiscretizer`
        - any instanciated model

        *estimator* allows the following values:

        - ``None``: the model is :epkg:`sklearn:linear_model:LinearRegression`
        - any instanciated model
        """
        BaseEstimator.__init__(self)
        if estimator is None:
            raise ValueError(  # pragma: no cover
                "estimator cannot be null.")
        if binner is None:
            raise TypeError(  # pragma: no cover
                "Unsupported options for binner=='tree' and model {}.".format(
                    type(estimator)))
        elif binner == "bins":
            binner = KBinsDiscretizer()
        self.binner = binner
        self.estimator = estimator
        self.n_jobs = n_jobs
        self.verbose = verbose

    @property
    def n_estimators_(self):
        """
        Returns the number of estimators = the number of buckets
        the data was split in.
        """
        return len(self.estimators_)

    def _mapping_train(self, X, binner):
        if hasattr(binner, "tree_"):
            tree = binner.tree_
            leaves = [i for i in range(len(tree.children_left))
                      if tree.children_left[i] <= i and tree.children_right[i] <= i]
            dec_path = self.binner_.decision_path(X)
            association = numpy.zeros((X.shape[0],))
            association[:] = -1
            mapping = {}
            ntree = 0
            for j in leaves:
                ind = dec_path[:, j] == 1
                ind = numpy.asarray(ind.todense()).flatten()
                if not numpy.any(ind):
                    # No training example for this bucket.
                    continue
                mapping[j] = ntree
                association[ind] = ntree
                ntree += 1

        elif hasattr(binner, "transform"):
            tr = binner.transform(X)
            unique = set()
            for x in tr:
                d = tuple(numpy.asarray(
                    x.todense()).ravel().astype(numpy.int32))
                unique.add(d)
            leaves = list(sorted(unique))
            association = numpy.zeros((X.shape[0],))
            association[:] = -1
            ntree = 0
            mapping = {}
            for i, le in enumerate(leaves):
                mapping[le] = i
            for i, x in enumerate(tr):
                d = tuple(numpy.asarray(
                    x.todense()).ravel().astype(numpy.int32))
                association[i] = mapping.get(d, -1)
        else:
            raise NotImplementedError(  # pragma: no cover
                "binner is not a decision tree or a transform")

        return association, mapping, leaves

    def transform_bins(self, X):
        """
        Maps every row to a tree in *self.estimators_*.
        """
        binner = self.binner_
        if hasattr(binner, "tree_"):
            dec_path = self.binner_.decision_path(X)
            association = numpy.zeros((X.shape[0],))
            association[:] = -1
            for j in self.leaves_:
                ind = dec_path[:, j] == 1
                ind = numpy.asarray(ind.todense()).flatten()
                if not numpy.any(ind):
                    # No training example for this bucket.
                    continue
                association[ind] = self.mapping_.get(j, -1)

        elif hasattr(binner, "transform"):
            association = numpy.zeros((X.shape[0],))
            association[:] = -1
            tr = binner.transform(X)
            for i, x in enumerate(tr):
                d = tuple(numpy.asarray(
                    x.todense()).ravel().astype(numpy.int32))
                association[i] = self.mapping_.get(d, -1)
        else:
            raise NotImplementedError(  # pragma: no cover
                "binner is not a decision tree or a transform")
        return association

    def fit(self, X, y, sample_weight=None):
        """
        Trains the binner and an estimator on every
        bucket.

        :param X: features, *X* is converted into an array if *X* is a dataframe
        :param y: target
        :param sample_weight: sample weights
        :return: self: returns an instance of self.

        Fitted attributes:

        * `binner_`: binner
        * `estimators_`: dictionary of estimators, each of them
            mapped to a leave to the tree
        * `mean_estimator_`: estimator trained on the whole
            datasets in case the binner can find a bucket for
            a new observation
        * `dim_`: dimension of the output
        * `mean_`: average targets
        """
        if isinstance(X, pandas.DataFrame):
            X = X.values
        if isinstance(X, list):
            raise TypeError(  # pragma: no cover
                "X cannot be a list.")
        binner = clone(self.binner)
        if sample_weight is None:
            self.binner_ = binner.fit(X, y)
        else:
            self.binner_ = binner.fit(X, y, sample_weight=sample_weight)

        association, self.mapping_, self.leaves_ = self._mapping_train(
            X, self.binner_)

        estimators = [clone(self.estimator) for i in self.mapping_]

        loop = (tqdm(range(len(estimators)))
                if self.verbose == 'tqdm' else range(len(estimators)))
        verbose = 1 if self.verbose == 'tqdm' else (1 if self.verbose else 0)

        self.mean_estimator_ = clone(self.estimator).fit(X, y, sample_weight)
        nb_classes = (None if not hasattr(self.mean_estimator_, 'classes_')
                      else len(set(self.mean_estimator_.classes_)))

        if hasattr(self, 'random_state') and self.random_state is not None:  # pylint: disable=E1101
            rnd = numpy.random.RandomState(  # pylint: disable=E1101
                self.random_state)  # pylint: disable=E1101
        else:
            rnd = None

        self.estimators_ = \
            Parallel(n_jobs=self.n_jobs, verbose=verbose,
                     **_joblib_parallel_args(prefer='threads'))(
                delayed(_fit_piecewise_estimator)(
                    i, estimators[i], X, y, sample_weight, association, nb_classes, rnd)
                for i in loop)

        self.dim_ = 1 if len(y.shape) == 1 else y.shape[1]
        if hasattr(self.estimators_[0], 'classes_'):
            self.classes_ = self.estimators_[0].classes_
        return self

    def _apply_predict_method(self, X, method, parallelized, dimout):
        """
        Generic *predict* method, works for *predict_proba* and
        *decision_function* as well.
        """
        if len(self.estimators_) == 0:
            raise RuntimeError(  # pragma: no cover
                "Estimator was apparently fitted but contains no estimator.")
        if not hasattr(self.estimators_[0], method):
            raise TypeError(  # pragma: no cover
                "Estimator {} does not have method '{}'.".format(
                    type(self.estimators_[0]), method))
        if isinstance(X, pandas.DataFrame):
            X = X.values

        association = self.transform_bins(X)

        indpred = Parallel(n_jobs=self.n_jobs, **_joblib_parallel_args(prefer='threads'))(
            delayed(parallelized)(i, model, X, association)
            for i, model in enumerate(self.estimators_))

        pred = numpy.zeros((X.shape[0], dimout)
                           if dimout > 1 else (X.shape[0],))
        indall = numpy.empty((X.shape[0],))
        indall[:] = False
        for ind, p in indpred:
            if ind is None:
                continue
            pred[ind] = p
            indall = numpy.logical_or(indall, ind)  # pylint: disable=E1111

        # no in a bucket
        indall = numpy.logical_not(indall)  # pylint: disable=E1111
        Xmissed = X[indall]
        if Xmissed.shape[0] > 0:
            meth = getattr(self.mean_estimator_, method)
            missed = meth(Xmissed)
            pred[indall] = missed
        return pred


class PiecewiseRegressor(PiecewiseEstimator, RegressorMixin):
    """
    Uses a :epkg:`decision tree` to split the space of features
    into buckets and trains a linear regression (default) on each of them.
    The second estimator is usually a :epkg:`sklearn:linear_model:LinearRegression`.
    It can also be :epkg:`sklearn:dummy:DummyRegressor` to just get
    the average on each bucket.
    """

    def __init__(self, binner=None, estimator=None, n_jobs=None, verbose=False):
        """
        @param      binner              transformer or predictor which creates the buckets
        @param      estimator           predictor trained on every bucket
        @param      n_jobs              number of parallel jobs (for training and predicting)
        @param      verbose             boolean or use ``'tqdm'`` to use :epkg:`tqdm`
                                        to fit the estimators

        *binner* allows the following values:

        - ``tree``: the model is :epkg:`sklearn:tree:DecisionTreeRegressor`
        - ``'bins'``: the model :epkg:`sklearn:preprocessing:KBinsDiscretizer`
        - any instanciated model

        *estimator* allows the following values:

        - ``None``: the model is :epkg:`sklearn:linear_model:LinearRegression`
        - any instanciated model
        """
        if estimator is None:
            estimator = LinearRegression()
        if binner in ('tree', None):
            binner = DecisionTreeRegressor(min_samples_leaf=2)
        RegressorMixin.__init__(self)
        PiecewiseEstimator.__init__(self, binner=binner, estimator=estimator,
                                    n_jobs=n_jobs, verbose=verbose)

    def predict(self, X):
        """
        Computes the predictions.

        :param X: features, *X* is converted into an array if *X* is a dataframe
        :return: predictions
        """
        return self._apply_predict_method(
            X, "predict", _predict_piecewise_estimator, self.dim_)


class PiecewiseClassifier(PiecewiseEstimator, ClassifierMixin):
    """
    Uses a :epkg:`decision tree` to split the space of features
    into buckets and trains a logistic regression (default) on each of them.
    The second estimator is usually a :epkg:`sklearn:linear_model:LogisticRegression`.
    It can also be :epkg:`sklearn:dummy:DummyClassifier` to just get
    the average on each bucket.

    The main issue with the *PiecewiseClassifier* is that each piece requires
    one example of each class in each bucket which may not happen.
    To avoid that, the training will pick up random example
    from other bucket to ensure this case does not happen.
    """

    def __init__(self, binner=None, estimator=None, n_jobs=None,
                 random_state=None, verbose=False):
        """
        @param      binner              transformer or predictor which creates the buckets
        @param      estimator           predictor trained on every bucket
        @param      n_jobs              number of parallel jobs (for training and predicting)
        @param      random_state        to pick up random examples when buckets do not
                                        contain enough examples of each class
        @param      verbose             boolean or use ``'tqdm'`` to use :epkg:`tqdm`
                                        to fit the estimators

        *binner* allows the following values:

        - ``tree``: the model is :epkg:`sklearn:tree:DecisionTreeClassifier`
        - ``'bins'``: the model :epkg:`sklearn:preprocessing:KBinsDiscretizer`
        - any instanciated model

        *estimator* allows the following values:

        - ``None``: the model is :epkg:`sklearn:linear_model:LogisticRegression`
        - any instanciated model
        """
        if estimator is None:
            estimator = LogisticRegression()
        if binner in ('tree', None):
            binner = DecisionTreeClassifier(min_samples_leaf=5)
        ClassifierMixin.__init__(self)
        PiecewiseEstimator.__init__(
            self, binner=binner, estimator=estimator,
            n_jobs=n_jobs, verbose=verbose)
        self.random_state = random_state

    def predict(self, X):
        """
        Computes the predictions.

        :param X: features, *X* is converted into an array if *X* is a dataframe
        :return: predictions
        """
        pred = self._apply_predict_method(
            X, "predict", _predict_piecewise_estimator, 1)
        return pred.astype(numpy.int32)

    def predict_proba(self, X):
        """
        Computes the predictions probabilities.

        :param X: features, *X* is converted into an array if *X* is a dataframe
        :return: predictions probabilities
        """
        return self._apply_predict_method(
            X, "predict_proba", _predict_proba_piecewise_estimator,
            len(self.mean_estimator_.classes_))

    def decision_function(self, X):
        """
        Computes the predictions probabilities.

        :param X: features, *X* is converted into an array if *X* is a dataframe
        :return: predictions probabilities
        """
        justone = self.mean_estimator_.decision_function(X[:1])
        return self._apply_predict_method(
            X, "decision_function", _decision_function_piecewise_estimator,
            1 if len(justone.shape) == 1 else justone.shape[1])



"""
@file
@brief Helpers to test a model which follows :epkg:`scikit-learn` API.
"""
import copy
import pickle
import pprint
from unittest import TestCase
from io import BytesIO
from numpy import ndarray
from numpy.testing import assert_almost_equal
from pandas.testing import assert_frame_equal
from sklearn.base import BaseEstimator
from sklearn.model_selection import train_test_split
from sklearn.base import clone
from sklearn.pipeline import make_pipeline
from sklearn.model_selection import GridSearchCV


def train_test_split_with_none(X, y=None, sample_weight=None, random_state=0):
    """
    Splits into train and test data even if they are None.

    @param      X               X
    @param      y               y
    @param      sample_weight   sample weight
    @param      random_state    random state
    @return                     similar to :epkg:`scikit-learn:model_selection:train_test_split`.
    """
    not_none = [_ for _ in [X, y, sample_weight] if _ is not None]
    res = train_test_split(*not_none)
    inc = len(not_none)
    trains = []
    tests = []
    for i in range(inc):
        trains.append(res[i * 2])
        tests.append(res[i * 2 + 1])
    while len(trains) < 3:
        trains.append(None)
        tests.append(None)
    X_train, y_train, w_train = trains
    X_test, y_test, w_test = tests
    return X_train, y_train, w_train, X_test, y_test, w_test


def test_sklearn_pickle(fct_model, X, y=None, sample_weight=None, **kwargs):
    """
    Creates a model, fit, predict and check the prediction
    are similar after the model was pickled, unpickled.

    @param      fct_model       function which creates the model
    @param      X               X
    @param      y               y
    @param      sample_weight   sample weight
    @param      kwargs          additional parameters for :epkg:`numpy:testing:assert_almost_equal`
    @return                     model, unpickled model

    :raises:
        AssertionError
    """
    X_train, y_train, w_train, X_test, _, __ = train_test_split_with_none(
        X, y, sample_weight)
    model = fct_model()
    if y_train is None and w_train is None:
        model.fit(X_train)
    else:
        try:
            model.fit(X_train, y_train, w_train)
        except TypeError:
            # Do not accept weights?
            model.fit(X_train, y_train)
    if hasattr(model, 'predict'):
        pred1 = model.predict(X_test)
    else:
        pred1 = model.transform(X_test)

    st = BytesIO()
    pickle.dump(model, st)
    data = BytesIO(st.getvalue())
    model2 = pickle.load(data)
    if hasattr(model2, 'predict'):
        pred2 = model2.predict(X_test)
    else:
        pred2 = model2.transform(X_test)
    if isinstance(pred1, ndarray):
        assert_almost_equal(pred1, pred2, **kwargs)
    else:
        assert_frame_equal(pred1, pred2, **kwargs)
    return model, model2


def _get_test_instance():
    try:
        from pyquickhelper.pycode import ExtTestCase  # pylint: disable=C0415
        cls = ExtTestCase
    except ImportError:

        class _ExtTestCase(TestCase):
            "simple test classe with a more methods"

            def assertIsInstance(self, inst, cltype):
                "checks that one instance is from one type"
                if not isinstance(inst, cltype):
                    raise AssertionError(
                        "Unexpected type {} != {}.".format(
                            type(inst), cltype))

        cls = _ExtTestCase
    return cls()


def test_sklearn_clone(fct_model, ext=None, copy_fitted=False):
    """
    Tests that a cloned model is similar to the original one.

    @param      fct_model       function which creates the model
    @param      ext             unit test class instance
    @param      copy_fitted     copy fitted parameters as well
    @return                     model, cloned model

    :raises:
        AssertionError
    """
    conv = fct_model()
    p1 = conv.get_params(deep=True)
    if copy_fitted:
        cloned = clone_with_fitted_parameters(conv)
    else:
        cloned = clone(conv)
    p2 = cloned.get_params(deep=True)
    if ext is None:
        ext = _get_test_instance()
    try:
        ext.assertEqual(set(p1), set(p2))
    except AssertionError as e:  # pragma no cover
        p1 = pprint.pformat(p1)
        p2 = pprint.pformat(p2)
        raise AssertionError(
            "Differences between\n----\n{0}\n----\n{1}".format(p1, p2)) from e

    for k in sorted(p1):
        if isinstance(p1[k], BaseEstimator) and isinstance(p2[k], BaseEstimator):
            if copy_fitted:
                assert_estimator_equal(p1[k], p2[k])
        elif isinstance(p1[k], list) and isinstance(p2[k], list):
            _assert_list_equal(p1[k], p2[k], ext)
        else:
            try:
                ext.assertEqual(p1[k], p2[k])
            except AssertionError as e:  # pragma no cover
                raise AssertionError(  # pylint: disable=W0707
                    "Difference for key '{0}'\n==1 {1}\n==2 {2}".format(
                        k, p1[k], p2[k]))
    return conv, cloned


def _assert_list_equal(l1, l2, ext):
    if len(l1) != len(l2):
        raise AssertionError(  # pragma no cover
            "Lists have different length {0} != {1}".format(len(l1), len(l2)))
    for a, b in zip(l1, l2):
        if isinstance(a, tuple) and isinstance(b, tuple):
            _assert_tuple_equal(a, b, ext)
        else:
            ext.assertEqual(a, b)


def _assert_dict_equal(a, b, ext):
    if not isinstance(a, dict):  # pragma no cover
        raise TypeError('a is not dict but {0}'.format(type(a)))
    if not isinstance(b, dict):  # pragma no cover
        raise TypeError('b is not dict but {0}'.format(type(b)))
    rows = []
    for key in sorted(b):
        if key not in a:
            rows.append("** Added key '{0}' in b".format(key))
        elif isinstance(a[key], BaseEstimator) and isinstance(b[key], BaseEstimator):
            assert_estimator_equal(a[key], b[key], ext)
        else:
            if a[key] != b[key]:
                rows.append(
                    "** Value != for key '{0}': != id({1}) != id({2})\n==1 {3}\n==2 {4}".format(
                        key, id(a[key]), id(b[key]), a[key], b[key]))
    for key in sorted(a):
        if key not in b:
            rows.append("** Removed key '{0}' in a".format(key))
    if len(rows) > 0:
        raise AssertionError(
            "Dictionaries are different\n{0}".format('\n'.join(rows)))


def _assert_tuple_equal(t1, t2, ext):
    if len(t1) != len(t2):  # pragma no cover
        raise AssertionError(
            "Lists have different length {0} != {1}".format(len(t1), len(t2)))
    for a, b in zip(t1, t2):
        if isinstance(a, BaseEstimator) and isinstance(b, BaseEstimator):
            assert_estimator_equal(a, b, ext)
        else:
            ext.assertEqual(a, b)


def assert_estimator_equal(esta, estb, ext=None):
    """
    Checks that two models are equal.

    @param      esta        first estimator
    @param      estb        second estimator
    @param      ext         unit test class

    The function raises an exception if the comparison fails.
    """
    if ext is None:
        ext = _get_test_instance()
    ext.assertIsInstance(esta, estb.__class__)
    ext.assertIsInstance(estb, esta.__class__)
    _assert_dict_equal(esta.get_params(), estb.get_params(), ext)
    for att in esta.__dict__:
        if (att.endswith('_') and not att.endswith('__')) or \
                (att.startswith('_') and not att.startswith('__')):
            if not hasattr(estb, att):  # pragma no cover
                raise AssertionError("Missing fitted attribute '{}' class {}\n==1 {}\n==2 {}".format(
                    att, esta.__class__, list(sorted(esta.__dict__)), list(sorted(estb.__dict__))))
            if isinstance(getattr(esta, att), BaseEstimator):
                assert_estimator_equal(
                    getattr(esta, att), getattr(estb, att), ext)
            else:
                ext.assertEqual(getattr(esta, att), getattr(estb, att))
    for att in estb.__dict__:
        if att.endswith('_') and not att.endswith('__'):
            if not hasattr(esta, att):  # pragma no cover
                raise AssertionError("Missing fitted attribute\n==1 {}\n==2 {}".format(
                    list(sorted(esta.__dict__)), list(sorted(estb.__dict__))))


def test_sklearn_grid_search_cv(fct_model, X, y=None, sample_weight=None, **grid_params):
    """
    Creates a model, checks that a grid search works with it.

    @param      fct_model       function which creates the model
    @param      X               X
    @param      y               y
    @param      sample_weight   sample weight
    @param      grid_params     parameter to use to run the grid search.
    @return                     dictionary with results

    :raises:
        AssertionError
    """
    X_train, y_train, w_train, X_test, y_test, w_test = train_test_split_with_none(
        X, y, sample_weight)
    model = fct_model()
    pipe = make_pipeline(model)
    name = model.__class__.__name__.lower()
    parameters = {name + "__" + k: v for k, v in grid_params.items()}
    if len(parameters) == 0:
        raise ValueError(
            "Some parameters must be tested when running grid search.")
    clf = GridSearchCV(pipe, parameters)
    if y_train is None and w_train is None:
        clf.fit(X_train)
    else:
        clf.fit(X_train, y_train, w_train)  # pylint: disable=E1121
    score = clf.score(X_test, y_test)
    ext = _get_test_instance()
    ext.assertIsInstance(score, float)
    return dict(model=clf, X_train=X_train, y_train=y_train, w_train=w_train,
                X_test=X_test, y_test=y_test, w_test=w_test, score=score)


def clone_with_fitted_parameters(est):
    """
    Clones an estimator with the fitted results.

    @param      est     estimator
    @return             cloned object
    """
    def adjust(obj1, obj2):
        if isinstance(obj1, list) and isinstance(obj2, list):
            for a, b in zip(obj1, obj2):
                adjust(a, b)
        elif isinstance(obj1, tuple) and isinstance(obj2, tuple):
            for a, b in zip(obj1, obj2):
                adjust(a, b)
        elif isinstance(obj1, dict) and isinstance(obj2, dict):
            for a, b in zip(obj1, obj2):
                adjust(obj1[a], obj2[b])
        elif isinstance(obj1, BaseEstimator) and isinstance(obj2, BaseEstimator):
            for k in obj1.__dict__:
                if hasattr(obj2, k):
                    v1 = getattr(obj1, k)
                    if callable(v1):
                        raise RuntimeError(
                            "Cannot migrate trained parameters for {}.".format(obj1))
                    elif isinstance(v1, BaseEstimator):
                        v1 = getattr(obj1, k)
                        setattr(obj2, k, clone_with_fitted_parameters(v1))
                    else:
                        adjust(getattr(obj1, k), getattr(obj2, k))
                elif (k.endswith('_') and not k.endswith('__')) or \
                     (k.startswith('_') and not k.startswith('__')):
                    v1 = getattr(obj1, k)
                    setattr(obj2, k, clone_with_fitted_parameters(v1))
                else:
                    raise RuntimeError(
                        "Cloned object is missing '{0}' in {1}.".format(k, obj2))

    if isinstance(est, BaseEstimator):
        cloned = clone(est)
        adjust(est, cloned)
        res = cloned
    elif isinstance(est, list):
        res = list(clone_with_fitted_parameters(o) for o in est)
    elif isinstance(est, tuple):
        res = tuple(clone_with_fitted_parameters(o) for o in est)
    elif isinstance(est, dict):
        res = {k: clone_with_fitted_parameters(v) for k, v in est.items()}
    else:
        res = copy.deepcopy(est)
    return res



# -*- coding: utf-8 -*-
"""
@file
@brief Implements a quantile non-linear regression.
"""
import inspect
import numpy as np
from sklearn.base import RegressorMixin
from sklearn.utils import check_X_y, column_or_1d
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.extmath import safe_sparse_dot
from sklearn.neural_network._base import DERIVATIVES, LOSS_FUNCTIONS
try:
    from sklearn.neural_network._multilayer_perceptron import BaseMultilayerPerceptron
except ImportError:  # pragma: no cover
    # scikit-learn < 0.22.
    from sklearn.neural_network.multilayer_perceptron import BaseMultilayerPerceptron
from sklearn.metrics import mean_absolute_error


def absolute_loss(y_true, y_pred):
    """
    Computes the absolute loss for regression.

    :param y_true: array-like or label indicator matrix
        Ground truth (correct) values.
    :param y_pred: array-like or label indicator matrix
        Predicted values, as returned by a regression estimator.
    :return: loss, float
        The degree to which the samples are correctly predicted.
    """
    return np.sum(np.abs(y_true - y_pred)) / y_true.shape[0]


def float_sign(a):
    "Returns 1 if *a > 0*, otherwise -1"
    if a > 1e-8:
        return 1.
    if a < -1e-8:
        return -1.
    return 0.


EXTENDED_LOSS_FUNCTIONS = {'absolute_loss': absolute_loss}
DERIVATIVE_LOSS_FUNCTIONS = {'absolute_loss': np.vectorize(float_sign)}


class CustomizedMultilayerPerceptron(BaseMultilayerPerceptron):
    """
    Customized MLP Perceptron based on
    `BaseMultilayerPerceptron
    <https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/neural_network/multilayer_perceptron.py#L40>`_.
    """

    def __init__(self, hidden_layer_sizes, activation, solver,
                 alpha, batch_size, learning_rate, learning_rate_init, power_t,
                 max_iter, loss, shuffle, random_state, tol, verbose,
                 warm_start, momentum, nesterovs_momentum, early_stopping,
                 validation_fraction, beta_1, beta_2, epsilon,
                 n_iter_no_change, max_fun):
        if 'max_fun' in inspect.signature(BaseMultilayerPerceptron.__init__).parameters:
            args = [15000]
        else:
            args = []
        BaseMultilayerPerceptron.__init__(  # pylint: disable=E1121
            self, hidden_layer_sizes, activation, solver, alpha, batch_size,
            learning_rate, learning_rate_init, power_t, max_iter, loss,
            shuffle, random_state, tol, verbose, warm_start, momentum,
            nesterovs_momentum, early_stopping, validation_fraction, beta_1, beta_2,
            epsilon, n_iter_no_change, *args)

    def _get_loss_function(self, loss_func_name):
        """
        Returns the loss functions.

        @param      loss_func_name      loss function name, see
                                        :epkg:`sklearn:neural_networks:MLPRegressor`
        """
        return LOSS_FUNCTIONS.get(loss_func_name, EXTENDED_LOSS_FUNCTIONS[loss_func_name])

    def _modify_loss_derivatives(self, last_deltas):
        """
        Modifies the loss derivatives.

        @param      last_deltas     last deltas is the difference between the output and the expected output
        @return                     modified derivatives
        """
        if self.loss == 'absolute_loss':
            return DERIVATIVE_LOSS_FUNCTIONS['absolute_loss'](last_deltas)
        return last_deltas  # pragma: no cover

    def _backprop(self, X, y, activations, deltas, coef_grads,
                  intercept_grads):
        """
        Computes the MLP loss function and its corresponding derivatives
        with respect to each parameter: weights and bias vectors.

        :param X: {array-like, sparse matrix}, shape (n_samples, n_features)
            The input data.
        :param y: array-like, shape (n_samples,)
            The target values.
        :param activations: list, length = n_layers - 1
             The ith element of the list holds the values of the ith layer.
        :param deltas: list, length = n_layers - 1
            The ith element of the list holds the difference between the
            activations of the i + 1 layer and the backpropagated error.
            More specifically, deltas are gradients of loss with respect to z
            in each layer, where z = wx + b is the value of a particular layer
            before passing through the activation function
        :param coef_grads: list, length = n_layers - 1
            The ith element contains the amount of change used to update the
            coefficient parameters of the ith layer in an iteration.
        :param intercept_grads: list, length = n_layers - 1
            The ith element contains the amount of change used to update the
            intercept parameters of the ith layer in an iteration.
        :return: loss, float
        :return: coef_grads, list, length = n_layers - 1
        :return: intercept_grads, list, length = n_layers - 1
        """
        n_samples = X.shape[0]

        # Forward propagate
        activations = self._forward_pass(activations)

        # Get loss
        loss_func_name = self.loss
        if loss_func_name == 'log_loss' and self.out_activation_ == 'logistic':
            loss_func_name = 'binary_log_loss'
        loss_function = self._get_loss_function(loss_func_name)
        loss = loss_function(y, activations[-1])
        # Add L2 regularization term to loss
        values = np.sum(
            np.array([np.dot(s.ravel(), s.ravel()) for s in self.coefs_]))
        loss += (0.5 * self.alpha) * values / n_samples

        # Backward propagate
        last = self.n_layers_ - 2

        # The calculation of delta[last] here works with following
        # combinations of output activation and loss function:
        # sigmoid and binary cross entropy, softmax and categorical cross
        # entropy, and identity with squared loss
        deltas[last] = activations[-1] - y

        # We insert the following modification to modify the gradient
        # due to the modification of the loss function.
        deltas[last] = self._modify_loss_derivatives(deltas[last])

        # Compute gradient for the last layer
        temp = self._compute_loss_grad(  # pylint: disable=E1111
            last, n_samples, activations, deltas, coef_grads, intercept_grads)
        if temp is None:
            # recent version of scikit-learn
            # Compute gradient for the last layer
            self._compute_loss_grad(
                last, n_samples, activations, deltas, coef_grads, intercept_grads)

            inplace_derivative = DERIVATIVES[self.activation]
            # Iterate over the hidden layers
            for i in range(self.n_layers_ - 2, 0, -1):
                deltas[i - 1] = safe_sparse_dot(deltas[i], self.coefs_[i].T)
                inplace_derivative(activations[i], deltas[i - 1])

                self._compute_loss_grad(
                    i - 1, n_samples, activations, deltas, coef_grads,
                    intercept_grads)
        else:  # pragma: no cover
            coef_grads, intercept_grads = temp

            # Iterate over the hidden layers
            for i in range(self.n_layers_ - 2, 0, -1):
                deltas[i - 1] = safe_sparse_dot(deltas[i], self.coefs_[i].T)
                inplace_derivative = DERIVATIVES[self.activation]
                inplace_derivative(activations[i], deltas[i - 1])

                coef_grads, intercept_grads = self._compute_loss_grad(  # pylint: disable=E1111
                    i - 1, n_samples, activations, deltas, coef_grads,
                    intercept_grads)

        return loss, coef_grads, intercept_grads


class QuantileMLPRegressor(CustomizedMultilayerPerceptron, RegressorMixin):
    """
    Quantile MLP Regression or neural networks regression
    trained with norm :epkg:`L1`. This class inherits from
    :epkg:`sklearn:neural_networks:MLPRegressor`.
    This model optimizes the absolute-loss using LBFGS or stochastic gradient
    descent. See @see cl CustomizedMultilayerPerceptron and
    @see fn absolute_loss.

    :param hidden_layer_sizes: tuple, length = n_layers - 2, default (100,)
        The ith element represents the number of neurons in the ith
        hidden layer.
    :param activation: {'identity', 'logistic', 'tanh', 'relu'}, default 'relu'
        Activation function for the hidden layer.
        - 'identity', no-op activation, useful to implement linear bottleneck,
          returns :math:`f(x) = x`
        - 'logistic', the logistic sigmoid function,
          returns :math:`f(x) = 1 / (1 + exp(-x))`.
        - 'tanh', the hyperbolic tan function,
          returns :math:`f(x) = tanh(x)`.
        - 'relu', the rectified linear unit function,
          returns :math:`f(x) = \\max(0, x)`.
    :param solver: ``{'lbfgs', 'sgd', 'adam'}``, default 'adam'
        The solver for weight optimization.
        - *'lbfgs'* is an optimizer in the family of quasi-Newton methods.
        - *'sgd'* refers to stochastic gradient descent.
        - *'adam'* refers to a stochastic gradient-based optimizer proposed by
          Kingma, Diederik, and Jimmy Ba
        Note: The default solver 'adam' works pretty well on relatively
        large datasets (with thousands of training samples or more) in terms of
        both training time and validation score.
        For small datasets, however, 'lbfgs' can converge faster and perform
        better.
    :param alpha: float, optional, default 0.0001
        :epkg:`L2` penalty (regularization term) parameter.
    :param batch_size: int, optional, default 'auto'
        Size of minibatches for stochastic optimizers.
        If the solver is 'lbfgs', the classifier will not use minibatch.
        When set to "auto", `batch_size=min(200, n_samples)`
    :param learning_rate: {'constant', 'invscaling', 'adaptive'}, default 'constant'
        Learning rate schedule for weight updates.
        - 'constant' is a constant learning rate given by
          'learning_rate_init'.
        - 'invscaling' gradually decreases the learning rate ``learning_rate_``
          at each time step 't' using an inverse scaling exponent of 'power_t'.
          effective_learning_rate = learning_rate_init / pow(t, power_t)
        - 'adaptive' keeps the learning rate constant to
          'learning_rate_init' as long as training loss keeps decreasing.
          Each time two consecutive epochs fail to decrease training loss by at
          least tol, or fail to increase validation score by at least tol if
          'early_stopping' is on, the current learning rate is divided by 5.
        Only used when solver='sgd'.
    :param learning_rate_init: double, optional, default 0.001
        The initial learning rate used. It controls the step-size
        in updating the weights. Only used when solver='sgd' or 'adam'.
    :param power_t: double, optional, default 0.5
        The exponent for inverse scaling learning rate.
        It is used in updating effective learning rate when the learning_rate
        is set to 'invscaling'. Only used when solver='sgd'.
    :param max_iter: int, optional, default 200
        Maximum number of iterations. The solver iterates until convergence
        (determined by 'tol') or this number of iterations. For stochastic
        solvers ('sgd', 'adam'), note that this determines the number of epochs
        (how many times each data point will be used), not the number of
        gradient steps.
    :param shuffle: bool, optional, default True
        Whether to shuffle samples in each iteration. Only used when
        solver='sgd' or 'adam'.
    :param random_state: int, RandomState instance or None, optional, default None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    :param tol: float, optional, default 1e-4
        Tolerance for the optimization. When the loss or score is not improving
        by at least ``tol`` for ``n_iter_no_change`` consecutive iterations,
        unless ``learning_rate`` is set to 'adaptive', convergence is
        considered to be reached and training stops.
    :param verbose: bool, optional, default False
        Whether to print progress messages to stdout.
    :param warm_start: bool, optional, default False
        When set to True, reuse the solution of the previous
        call to fit as initialization, otherwise, just erase the
        previous solution. See :term:`the Glossary <warm_start>`.
    :param momentum: float, default 0.9
        Momentum for gradient descent update.  Should be between 0 and 1. Only
        used when solver='sgd'.
    :param nesterovs_momentum: boolean, default True
        Whether to use Nesterov's momentum. Only used when solver='sgd' and
        momentum > 0.
    :param early_stopping: bool, default False
        Whether to use early stopping to terminate training when validation
        score is not improving. If set to true, it will automatically set
        aside 10% of training data as validation and terminate training when
        validation score is not improving by at least ``tol`` for
        ``n_iter_no_change`` consecutive epochs.
        Only effective when solver='sgd' or 'adam'
    :param validation_fraction: float, optional, default 0.1
        The proportion of training data to set aside as validation set for
        early stopping. Must be between 0 and 1.
        Only used if early_stopping is True
    :param beta_1: float, optional, default 0.9
        Exponential decay rate for estimates of first moment vector in adam,
        should be in [0, 1). Only used when solver='adam'
    :param beta_2: float, optional, default 0.999
        Exponential decay rate for estimates of second moment vector in adam,
        should be in [0, 1). Only used when solver='adam'
    :param epsilon: float, optional, default 1e-8
        Value for numerical stability in adam. Only used when solver='adam'
    :param n_iter_no_change: int, optional, default 10
        Maximum number of epochs to not meet ``tol`` improvement.
        Only effective when solver='sgd' or 'adam'

    Fitted attributes:

    * `loss_`: float
        The current loss computed with the loss function.
    * `coefs_`: list, length n_layers - 1
        The ith element in the list represents the weight matrix corresponding
        to layer i.
    * `intercepts_`: list, length n_layers - 1
        The ith element in the list represents the bias vector corresponding to
        layer i + 1.
    * `n_iter_`: int,
        The number of iterations the solver has ran.
    * `n_layers_`: int
        Number of layers.
    * `n_outputs_`: int
        Number of outputs.
    * `out_activation_`: string
        Name of the output activation function.
    """

    def __init__(self,
                 hidden_layer_sizes=(100,), activation="relu",
                 solver='adam', alpha=0.0001,
                 batch_size='auto', learning_rate="constant",
                 learning_rate_init=0.001,
                 power_t=0.5, max_iter=200, shuffle=True,
                 random_state=None, tol=1e-4,
                 verbose=False, warm_start=False, momentum=0.9,
                 nesterovs_momentum=True, early_stopping=False,
                 validation_fraction=0.1, beta_1=0.9, beta_2=0.999,
                 epsilon=1e-8, n_iter_no_change=10,
                 **kwargs):
        """
        See :epkg:`sklearn:neural_networks:MLPRegressor`
        """
        sup = super(QuantileMLPRegressor, self)  # pylint: disable=R1725
        if "max_fun" not in kwargs:
            sig = inspect.signature(sup.__init__)
            if "max_fun" in sig.parameters:
                kwargs['max_fun'] = 15000
        sup.__init__(hidden_layer_sizes=hidden_layer_sizes,
                     activation=activation, solver=solver, alpha=alpha,
                     batch_size=batch_size, learning_rate=learning_rate,
                     learning_rate_init=learning_rate_init, power_t=power_t,
                     max_iter=max_iter, loss='absolute_loss', shuffle=shuffle,
                     random_state=random_state, tol=tol, verbose=verbose,
                     warm_start=warm_start, momentum=momentum,
                     nesterovs_momentum=nesterovs_momentum,
                     early_stopping=early_stopping,
                     validation_fraction=validation_fraction,
                     beta_1=beta_1, beta_2=beta_2, epsilon=epsilon,
                     n_iter_no_change=n_iter_no_change, **kwargs)

    def predict(self, X):
        """
        Predicts using the multi-layer perceptron model.

        :param X: {array-like, sparse matrix}, shape (n_samples, n_features)
            The input data.
        :return: y : array-like, shape (n_samples, n_outputs)
            The predicted values.
        """
        check_is_fitted(self)
        if hasattr(self, '_predict'):
            y_pred = self._predict(X)
        else:
            y_pred = self._forward_pass_fast(X)
        if y_pred.shape[1] == 1:
            return y_pred.ravel()
        return y_pred

    def _validate_input(self, X, y, incremental, reset=False):
        X, y = check_X_y(X, y, accept_sparse=['csr', 'csc', 'coo'],
                         multi_output=True, y_numeric=True)
        if y.ndim == 2 and y.shape[1] == 1:
            y = column_or_1d(y, warn=True)
        return X, y

    def score(self, X, y, sample_weight=None):
        """
        Returns mean absolute error regression loss.

        :param X: array-like, shape = (n_samples, n_features)
            Test samples.
        :param y: array-like, shape = (n_samples) or (n_samples, n_outputs)
            True values for X.
        :param sample_weight: array-like, shape = [n_samples], optional
            Sample weights.
        :return: score, float
            mean absolute error regression loss
        """
        pred = self.predict(X)
        return mean_absolute_error(y, pred, sample_weight=sample_weight)



"""
@file
@brief Implements a way to get close examples based
on the output of a machine learned model.
"""
import json
import zipfile
import pandas
import numpy
from sklearn.neighbors import NearestNeighbors
from pandas_streaming.df import to_zip, read_zip
from ..helpers.parameters import format_function_call


class SearchEngineVectors:
    """
    Implements a kind of local search engine which
    looks for similar results assuming they are vectors.
    The class is using
    :epkg:`sklearn:neighborsNearestNeighbors` to find
    the nearest neighbors of a vector and follows
    the same API.
    The class populates members:

    * ``features_``: vectors used to compute the neighbors
    * ``knn_``: parameters for the :epkg:`sklearn:neighborsNearestNeighbors`
    * ``metadata_``: metadata, can be None
    """

    def __init__(self, **pknn):
        """
        @param      pknn        list of parameters, see :epkg:`sklearn:neighborsNearestNeighbors`
        """
        self.pknn = pknn

    def __repr__(self):
        """
        usual
        """
        return format_function_call(self.__class__.__name__, self.pknn)

    def _is_iterable(self, data):
        """
        Tells if an objet is an iterator or not.
        """
        try:
            iter(data)
            return not isinstance(data, (list, tuple, pandas.DataFrame, numpy.ndarray))
        except TypeError:
            return False

    def _prepare_fit(self, data=None, features=None, metadata=None, transform=None):
        """
        Stores data in the class itself.

        @param      data        a :epkg:`dataframe` or None if the
                                the features and the metadata
                                are specified with an array and a
                                dictionary
        @param      features    features columns or an array
        @param      metadata    data
        @param      transform   transform each vector before using it

        *transform* is a function whose signature::

            def transform(vec, many):
                # Many tells is the functions receives many vectors
                # or just one (many=False).

        Function *transform* is applied only if
        *data* is not None.
        """
        iterate = self._is_iterable(data)
        if iterate:
            if data is None:
                raise ValueError(  # pragma: no cover
                    "iterator is True, data must be specified.")
            if features is not None:
                raise ValueError(  # pragma: no cover
                    "iterator is True, features must be None.")
            if metadata is not None:
                raise ValueError(  # pragma: no cover
                    "iterator is True, metadata must be None.")
            metas = []
            arrays = []
            for row in data:
                if not isinstance(row, tuple):
                    raise TypeError(  # pragma: no cover
                        'data must be an iterator on tuple')
                if len(row) != 2:
                    raise ValueError(  # pragma: no cover
                        'data must be an iterator on tuple on two elements')
                arr, meta = row
                if not isinstance(meta, dict):
                    raise TypeError(  # pragma: no cover
                        'Second element of the tuple must be a dictionary')
                metas.append(meta)
                if transform is None:
                    tradd = arr
                else:
                    tradd = transform(arr, False)
                if not isinstance(tradd, numpy.ndarray):
                    if transform is None:
                        raise TypeError(  # pragma: no cover
                            "feature should be of type numpy.array not {}".format(type(tradd)))
                    else:
                        raise TypeError(  # pragma: no cover
                            "output of method transform ({}) should be of type numpy.array not {}".format(
                                transform, type(tradd)))
                arrays.append(tradd)
            self.features_ = numpy.vstack(arrays)
            self.metadata_ = pandas.DataFrame(metas)
        elif data is None:
            if not isinstance(features, numpy.ndarray):
                raise TypeError(  # pragma: no cover
                    "features must be an array if data is None")
            self.features_ = features
            self.metadata_ = metadata
        else:
            if not isinstance(data, pandas.DataFrame):
                raise ValueError(  # pragma: no cover
                    "data should be a dataframe")
            self.features_ = data[features]
            self.metadata_ = data[metadata] if metadata else None

    def fit(self, data=None, features=None, metadata=None):
        """
        Every vector comes with a list of metadata.

        @param      data        a dataframe or None if the
                                the features and the metadata
                                are specified with an array and a
                                dictionary
        @param      features    features columns or an array
        @param      metadata    data
        """
        self._prepare_fit(data=data, features=features, metadata=metadata)
        return self._fit_knn()

    def _fit_knn(self):
        """
        Fits the nearest neighbors.
        """
        self.knn_ = NearestNeighbors(**self.pknn)
        self.knn_.fit(self.features_)
        return self

    def _first_pass(self, X, n_neighbors=None):
        """
        Finds the closest *n_neighbors*.

        @param      X               features
        @param      n_neighbors     number of neighbors to get (default is the value passed to the constructor)
        @return                     *dist*, *ind*

        *dist* is an array representing the lengths to points,
        *ind* contains the indices of the nearest points in the population matrix.
        """
        if isinstance(X, list):
            if len(X) == 0 or isinstance(X[0], (list, tuple)):
                raise TypeError(  # pragma: no cover
                    "X must be a list or a vector (1)")
            X = [X]
        if isinstance(X, numpy.ndarray) and (len(X.shape) > 1 and X.shape[0] != 1):
            raise TypeError(  # pragma: no cover
                "X must be a list or a vector (2)")
        dist, ind = self.knn_.kneighbors(
            X, n_neighbors=n_neighbors, return_distance=True)
        ind = ind.ravel()
        dist = dist.ravel()
        return dist, ind

    def _second_pass(self, X, dist, ind):
        """
        Reorders the closest *n_neighbors*.

        @param      X               features
        @param      dist            array representing the lengths to points
        @param      ind             indices of the nearest points in the population matrix
        @return                     *score*, *ind*

        *score* is an array representing the lengths to points,
        *ind* contains the indices of the nearest points in the population matrix.
        """
        return dist, ind

    def kneighbors(self, X, n_neighbors=None):
        """
        Searches for neighbors close to *X*.

        @param      X               features
        @return                     score, ind, meta

        *score* is an array representing the lengths to points,
        *ind* contains the indices of the nearest points in the population matrix,
        *meta* is the metadata
        """
        dist, ind = self._first_pass(X, n_neighbors=n_neighbors)
        score, ind = self._second_pass(X, dist, ind)
        rind = ind
        if self.metadata_ is None:
            rmeta = None
        elif hasattr(self.metadata_, 'iloc'):
            rmeta = self.metadata_.iloc[ind, :]
        elif len(self.metadata_.shape) == 1:
            rmeta = self.metadata_[ind]
        else:
            rmeta = self.metadata_[ind, :]
        return score, rind, rmeta

    def to_zip(self, zipfilename, **kwargs):
        """
        Saves the features and the metadata into a zipfile.
        The function does not save the *k-nn*.

        @param      zipfilename a :epkg:`*py:zipfile:ZipFile` or a filename
        @param      kwargs      parameters for :epkg:`pandas:to_csv` (for the metadata)
        @return                 zipfilename

        The function relies on function
        `to_zip <http://www.xavierdupre.fr/app/pandas_streaming/helpsphinx/pandas_streaming/df/
        dataframe_io.html#pandas_streaming.df.dataframe_io.to_zip>`_.
        It only works for :epkg:`Python` 3.6+.
        """
        if isinstance(zipfilename, str):
            zf = zipfile.ZipFile(zipfilename, 'w')
            close = True
        else:
            zf = zipfilename
            close = False
        if 'index' is not kwargs:
            kwargs['index'] = False
        to_zip(self.features_, zf, 'SearchEngineVectors-features.npy')
        to_zip(self.metadata_, zf, 'SearchEngineVectors-metadata.csv', **kwargs)
        js = json.dumps(self.pknn)
        zf.writestr('SearchEngineVectors-knn.json', js)
        if close:
            zf.close()

    @staticmethod
    def read_zip(zipfilename, **kwargs):
        """
        Restore the features, the metadata to a @see cl SearchEngineVectors.

        @param      zipfilename a :epkg:`*py:zipfile:ZipFile` or a filename
        @param      zname       a filename in th zipfile
        @param      kwargs      parameters for :epkg:`pandas:read_csv`
        @return                 @see cl SearchEngineVectors

        It only works for :epkg:`Python` 3.6+.
        """
        if isinstance(zipfilename, str):
            zf = zipfile.ZipFile(zipfilename, 'r')
            close = True
        else:
            zf = zipfilename
            close = False
        feat = read_zip(zf, 'SearchEngineVectors-features.npy')
        meta = read_zip(zf, 'SearchEngineVectors-metadata.csv', **kwargs)
        js = zf.read('SearchEngineVectors-knn.json')
        knn = json.loads(js)
        if close:
            zf.close()

        obj = SearchEngineVectors(**knn)
        obj.fit(features=feat, metadata=meta)
        return obj



# -*- coding: utf-8 -*-
"""
@file
@brief Implements a quantile linear regression.
"""
import numpy
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_absolute_error


class QuantileLinearRegression(LinearRegression):
    """
    Quantile Linear Regression or linear regression
    trained with norm :epkg:`L1`. This class inherits from
    :epkg:`sklearn:linear_models:LinearRegression`.
    See notebook :ref:`quantileregressionrst`.

    Norm :epkg:`L1` is chosen if ``quantile=0.5``, otherwise,
    for *quantile=*:math:`\\rho`,
    the following error is optimized:

    .. math::

        \\sum_i \\rho |f(X_i) - Y_i|^- + (1-\\rho) |f(X_i) - Y_i|^+

    where :math:`|f(X_i) - Y_i|^-= \\max(Y_i - f(X_i), 0)` and
    :math:`|f(X_i) - Y_i|^+= \\max(f(X_i) - Y_i, 0)`.
    :math:`f(i)` is the prediction, :math:`Y_i` the expected
    value.
    """

    def __init__(self, fit_intercept=True, normalize=False, copy_X=True,
                 n_jobs=1, delta=0.0001, max_iter=10, quantile=0.5,
                 verbose=False):
        """
        :param fit_intercept: boolean, optional, default True
            whether to calculate the intercept for this model. If set
            to False, no intercept will be used in calculations
            (e.g. data is expected to be already centered).
        :param normalize: boolean, optional, default False
            This parameter is ignored when ``fit_intercept`` is set to False.
            If True, the regressors X will be normalized before regression by
            subtracting the mean and dividing by the l2-norm.
            If you wish to standardize, please use
            :class:`sklearn.preprocessing.StandardScaler` before calling ``fit`` on
            an estimator with ``normalize=False``.
        :param copy_X: boolean, optional, default True
            If True, X will be copied; else, it may be overwritten.
        :param n_jobs: int, optional, default 1
            The number of jobs to use for the computation.
            If -1 all CPUs are used. This will only provide speedup for
            n_targets > 1 and sufficient large problems.
        :param max_iter: int, optional, default 1
            The number of iteration to do at training time.
            This parameter is specific to the quantile regression.
        :param delta: float, optional, default 0.0001
            Used to ensure matrices has an inverse
            (*M + delta*I*).
        :param quantile: float, by default 0.5,
            determines which quantile to use
            to estimate the regression.
        :param verbose: bool, optional, default False
            Prints error at each iteration of the optimisation.
        """
        LinearRegression.__init__(self, fit_intercept=fit_intercept, normalize=normalize,
                                  copy_X=copy_X, n_jobs=n_jobs)
        self.max_iter = max_iter
        self.verbose = verbose
        self.delta = delta
        self.quantile = quantile

    def fit(self, X, y, sample_weight=None):
        """
        Fits a linear model with :epkg:`L1` norm which
        is equivalent to a quantile regression.
        The implementation is not the most efficient
        as it calls multiple times method fit
        from :epkg:`sklearn:linear_models:LinearRegression`.
        Data gets checked and rescaled each time.
        The optimization follows the algorithm
        `Iteratively reweighted least squares
        <https://en.wikipedia.org/wiki/Iteratively_reweighted_least_squares>`_.
        It is described in French at
        `Régression quantile
        <http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx3/notebooks/td_note_2017_2.html>`_.

        :param X: numpy array or sparse matrix of shape [n_samples,n_features]
            Training data
        :param y: numpy array of shape [n_samples, n_targets]
            Target values. Will be cast to X's dtype if necessary
        :param sample_weight: numpy array of shape [n_samples]
            Individual weights for each sample
        :return: self, returns an instance of self.

        Fitted attributes:

        * `coef_`: array, shape (n_features, ) or (n_targets, n_features)
            Estimated coefficients for the linear regression problem.
            If multiple targets are passed during the fit (y 2D), this
            is a 2D array of shape (n_targets, n_features), while if only
            one target is passed, this is a 1D array of length n_features.
        *  `intercept_`: array
            Independent term in the linear model.
        * `n_iter_`: int
            Number of iterations at training time.
        """
        if len(y.shape) > 1 and y.shape[1] != 1:
            raise ValueError("QuantileLinearRegression only works for Y real")

        def compute_z(Xm, beta, Y, W, delta=0.0001):
            "compute z"
            deltas = numpy.ones(X.shape[0]) * delta
            epsilon, mult = QuantileLinearRegression._epsilon(
                Y, Xm @ beta, self.quantile)
            r = numpy.reciprocal(numpy.maximum(  # pylint: disable=E1111
                epsilon, deltas))  # pylint: disable=E1111
            if mult is not None:
                epsilon *= 1 - mult
                r *= 1 - mult
            return r, epsilon

        if not isinstance(X, numpy.ndarray):
            if hasattr(X, 'values'):
                X = X.values
            else:
                raise TypeError("X must be an array or a dataframe.")

        if self.fit_intercept:
            Xm = numpy.hstack([X, numpy.ones((X.shape[0], 1))])
        else:
            Xm = X

        clr = LinearRegression(fit_intercept=False, copy_X=self.copy_X,
                               n_jobs=self.n_jobs, normalize=self.normalize)

        W = numpy.ones(X.shape[0]) if sample_weight is None else sample_weight
        self.n_iter_ = 0
        lastE = None
        for i in range(0, self.max_iter):
            clr.fit(Xm, y, W)
            beta = clr.coef_
            W, epsilon = compute_z(Xm, beta, y, W, delta=self.delta)
            if sample_weight is not None:
                W *= sample_weight
                epsilon *= sample_weight
            E = epsilon.sum()
            self.n_iter_ = i
            if self.verbose:
                print(  # pragma: no cover
                    '[QuantileLinearRegression.fit] iter={0} error={1}'.format(i + 1, E))
            if lastE is not None and lastE == E:
                break
            lastE = E

        if self.fit_intercept:
            self.coef_ = beta[:-1]
            self.intercept_ = beta[-1]
        else:
            self.coef_ = beta
            self.intercept_ = 0

        return self

    @staticmethod
    def _epsilon(y_true, y_pred, quantile, sample_weight=None):
        diff = y_pred - y_true
        epsilon = numpy.abs(diff)
        if quantile != 0.5:
            sign = numpy.sign(diff)  # pylint: disable=E1111
            mult = numpy.ones(y_true.shape[0])
            mult[sign > 0] *= quantile  # pylint: disable=W0143
            mult[sign < 0] *= (1 - quantile)  # pylint: disable=W0143
        else:
            mult = None
        if sample_weight is not None:
            epsilon *= sample_weight
        return epsilon, mult

    def score(self, X, y, sample_weight=None):
        """
        Returns Mean absolute error regression loss.

        :param X: array-like, shape = (n_samples, n_features)
            Test samples.
        :param y: array-like, shape = (n_samples) or (n_samples, n_outputs)
            True values for X.
        :param sample_weight: array-like, shape = [n_samples], optional
            Sample weights.
        :return: score : float
            mean absolute error regression loss
        """
        pred = self.predict(X)

        if self.quantile != 0.5:
            epsilon, mult = QuantileLinearRegression._epsilon(
                y, pred, self.quantile, sample_weight)
            if mult is not None:
                epsilon *= mult * 2
            return epsilon.sum() / X.shape[0]
        else:
            return mean_absolute_error(y, pred, sample_weight=sample_weight)



"""
@file
@brief Implements a predicatable *t-SNE*.
"""
import inspect
from sklearn.base import BaseEstimator, TransformerMixin, clone
from sklearn.manifold import TSNE
from sklearn.neural_network import MLPRegressor
from sklearn.metrics import mean_squared_error


class PredictableTSNE(BaseEstimator, TransformerMixin):
    """
    :epkg:`t-SNE` is an interesting
    transform which can only be used to study data as there is no
    way to reproduce the result once it was fitted. That's why
    the class :epkg:`TSNE` does not have any method *transform*, only
    `fit_transform <https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html#sklearn.manifold.TSNE.fit_transform>`_.
    This example proposes a way to train a machine learned model
    which approximates the outputs of a :epkg:`TSNE` transformer.
    Notebooks :ref:`predictabletsnerst` gives an example on how to
    use this class.
    """

    def __init__(self, normalizer=None, transformer=None, estimator=None,
                 normalize=True, keep_tsne_outputs=False):
        """
        @param      normalizer          None by default
        @param      transformer         :epkg:`sklearn:manifold:TSNE`
                                        by default
        @param      estimator           :epkg:`sklearn:neural_network:MLPRegressor`
                                        by default
        @param      normalize           normalizes the outputs, centers and normalizes
                                        the output of the *t-SNE* and applies that same
                                        normalization to he prediction of the estimator
        @param      keep_tsne_output    if True, keep raw outputs of
                                        :epkg:`TSNE` is stored in member
                                        *tsne_outputs_*
        """
        TransformerMixin.__init__(self)
        BaseEstimator.__init__(self)
        if estimator is None:
            estimator = MLPRegressor()
        if transformer is None:
            transformer = TSNE()
        self.estimator = estimator
        self.transformer = transformer
        self.normalizer = normalizer
        self.keep_tsne_outputs = keep_tsne_outputs
        if normalizer is not None and not hasattr(normalizer, "transform"):
            raise AttributeError(  # pragma: no cover
                "normalizer {} does not have a 'transform' method.".format(
                    type(normalizer)))
        if not hasattr(transformer, "fit_transform"):
            raise AttributeError(  # pragma: no cover
                "transformer {} does not have a 'fit_transform' method.".format(
                    type(transformer)))
        if not hasattr(estimator, "predict"):
            raise AttributeError(  # pragma: no cover
                "estimator {} does not have a 'predict' method.".format(
                    type(estimator)))
        self.normalize = normalize

    def fit(self, X, y, sample_weight=None):
        """
        Trains a :epkg:`TSNE` then trains an estimator
        to approximate its outputs.

        :param X: numpy array or sparse matrix of shape [n_samples,n_features]
            Training data
        :param y: numpy array of shape [n_samples, n_targets]
            Target values. Will be cast to X's dtype if necessary
        :param sample_weight: numpy array of shape [n_samples]
            Individual weights for each sample
        :return: self, returns an instance of self.

        Fitted attributes:

        * `normalizer_`: trained normalier
        * `transformer_`: trained transformeer
        * `estimator_`: trained regressor
        * `tsne_outputs_`: t-SNE outputs if *keep_tsne_outputs* is True
        * `mean_`: average of the *t-SNE* output on each dimension
        * `inv_std_`: inverse of the standard deviation of the *t-SNE*
            output on each dimension
        * `loss_`: loss (:epkg:`sklearn:metrics:mean_squared_error`) between the predictions
            and the outputs of t-SNE
        """
        params = dict(y=y, sample_weight=sample_weight)

        if self.normalizer is not None:
            sig = inspect.signature(self.normalizer.transform)
            pars = {}
            for p in ['sample_weight', 'y']:
                if p in sig.parameters and p in params:
                    pars[p] = params[p]
            self.normalizer_ = clone(self.normalizer).fit(X, **pars)
            X = self.normalizer_.transform(X)
        else:
            self.normalizer_ = None

        self.transformer_ = clone(self.transformer)

        sig = inspect.signature(self.transformer.fit_transform)
        pars = {}
        for p in ['sample_weight', 'y']:
            if p in sig.parameters and p in params:
                pars[p] = params[p]
        target = self.transformer_.fit_transform(X, **pars)

        sig = inspect.signature(self.estimator.fit)
        if 'sample_weight' in sig.parameters:
            self.estimator_ = clone(self.estimator).fit(
                X, target, sample_weight=sample_weight)
        else:
            self.estimator_ = clone(self.estimator).fit(X, target)
        mean = target.mean(axis=0)
        var = target.std(axis=0)
        self.mean_ = mean
        self.inv_std_ = 1. / var
        exp = (target - mean) * self.inv_std_
        got = (self.estimator_.predict(X) - mean) * self.inv_std_
        self.loss_ = mean_squared_error(exp, got)
        if self.keep_tsne_outputs:
            self.tsne_outputs_ = exp if self.normalize else target
        return self

    def transform(self, X):
        """
        Runs the predictions.

        :param X: numpy array or sparse matrix of shape [n_samples,n_features]
            Training data
        :return: tranformed *X*
        """
        if self.normalizer_ is not None:
            X = self.normalizer_.transform(X)
        pred = self.estimator_.predict(X)
        if self.normalize:
            pred -= self.mean_
            pred *= self.inv_std_
        return pred



# pylint: disable=C0302
"""
@file
@brief Implements k-means with norms L1 and L2.
"""
import warnings
import numpy
from scipy.sparse import issparse
from joblib import Parallel, delayed, effective_n_jobs
from sklearn.cluster import KMeans
from sklearn.cluster._kmeans import _tolerance as _tolerance_skl
from sklearn.exceptions import ConvergenceWarning
from sklearn.metrics.pairwise import (
    euclidean_distances, manhattan_distances,
    pairwise_distances_argmin_min)
from sklearn.utils import check_random_state, check_array
from sklearn.utils.validation import _num_samples, check_is_fitted
from sklearn.utils.extmath import stable_cumsum
try:
    from sklearn.cluster._kmeans import _check_sample_weight
except ImportError:
    from sklearn.cluster._kmeans import (
        _check_normalize_sample_weight as _check_sample_weight)
from ._kmeans_022 import (
    _labels_inertia_skl,
    _labels_inertia_precompute_dense)


def _k_init(norm, X, n_clusters, random_state, n_local_trials=None):
    """Init n_clusters seeds according to k-means++

    :param norm: `l1` or `l2`
        manhattan or euclidean distance
    :param X: array or sparse matrix, shape (n_samples, n_features)
        The data to pick seeds for. To avoid memory copy, the input data
        should be double precision (dtype=numpy.float64).
    :param n_clusters: integer
        The number of seeds to choose
    :param random_state: int, RandomState instance
        The generator used to initialize the centers. Use an int to make the
        randomness deterministic.
        See :term:`Glossary <random_state>`.
    :param n_local_trials: integer, optional
        The number of seeding trials for each center (except the first),
        of which the one reducing inertia the most is greedily chosen.
        Set to None to make the number of trials depend logarithmically
        on the number of seeds (2+log(k)); this is the default.
    """
    n_samples, n_features = X.shape

    centers = numpy.empty((n_clusters, n_features), dtype=X.dtype)

    # Set the number of local seeding trials if none is given
    if n_local_trials is None:
        # This is what Arthur/Vassilvitskii tried, but did not report
        # specific results for other than mentioning in the conclusion
        # that it helped.
        n_local_trials = 2 + int(numpy.log(n_clusters))

    # Pick first center randomly
    center_id = random_state.randint(n_samples)
    if issparse(X):
        centers[0] = X[center_id].toarray()
    else:
        centers[0] = X[center_id]

    # Initialize list of closest distances and calculate current potential
    if norm.lower() == 'l2':
        dist_fct = lambda x, y: euclidean_distances(x, y, squared=True)
    elif norm.lower() == 'l1':
        dist_fct = lambda x, y: manhattan_distances(x, y)
    else:
        raise NotImplementedError(  # pragma no cover
            "norm must be 'l1' or 'l2' not '{}'.".format(norm))

    closest_dist_sq = dist_fct(centers[0, numpy.newaxis], X)
    current_pot = closest_dist_sq.sum()

    # Pick the remaining n_clusters-1 points
    for c in range(1, n_clusters):
        # Choose center candidates by sampling with probability proportional
        # to the squared distance to the closest existing center
        rand_vals = random_state.random_sample(n_local_trials) * current_pot
        candidate_ids = numpy.searchsorted(stable_cumsum(closest_dist_sq),
                                           rand_vals)
        numpy.clip(candidate_ids, None, closest_dist_sq.size - 1,
                   out=candidate_ids)

        # Compute distances to center candidates
        distance_to_candidates = dist_fct(X[candidate_ids], X)

        # update closest distances squared and potential for each candidate
        numpy.minimum(closest_dist_sq, distance_to_candidates,
                      out=distance_to_candidates)
        candidates_pot = distance_to_candidates.sum(axis=1)

        # Decide which candidate is the best
        best_candidate = numpy.argmin(candidates_pot)
        current_pot = candidates_pot[best_candidate]
        closest_dist_sq = distance_to_candidates[best_candidate]
        best_candidate = candidate_ids[best_candidate]

        # Permanently add best center candidate found in local tries
        if issparse(X):
            centers[c] = X[best_candidate].toarray()
        else:
            centers[c] = X[best_candidate]

    return centers


def _init_centroids(norm, X, k, init, random_state=None,
                    init_size=None):
    """Compute the initial centroids

    :param norm: 'l1' or 'l2'
    :param X: array, shape (n_samples, n_features)
    :param k: int
        number of centroids
    :param init: {'k-means++', 'random' or ndarray or callable} optional
        Method for initialization
    :param random_state: int, RandomState instance or None (default)
        Determines random number generation for centroid initialization. Use
        an int to make the randomness deterministic.
        See :term:`Glossary <random_state>`.
    :param init_size: int, optional
        Number of samples to randomly sample for speeding up the
        initialization (sometimes at the expense of accuracy): the
        only algorithm is initialized by running a batch KMeans on a
        random subset of the data. This needs to be larger than k.
    :return: centers, array, shape(k, n_features)
    """
    random_state = check_random_state(random_state)
    n_samples = X.shape[0]

    if init_size is not None and init_size < n_samples:
        if init_size < k:  # pragma: no cover
            warnings.warn(
                "init_size=%d should be larger than k=%d. "
                "Setting it to 3*k" % (init_size, k),
                RuntimeWarning, stacklevel=2)
            init_size = 3 * k
        init_indices = random_state.randint(0, n_samples, init_size)
        X = X[init_indices]
        n_samples = X.shape[0]
    elif n_samples < k:
        raise ValueError(
            "n_samples=%d should be larger than k=%d" % (n_samples, k))

    if isinstance(init, str) and init == 'k-means++':
        centers = _k_init(norm, X, k, random_state=random_state)
    elif isinstance(init, str) and init == 'random':
        seeds = random_state.permutation(n_samples)[:k]
        centers = X[seeds]
    elif hasattr(init, '__array__'):
        # ensure that the centers have the same dtype as X
        # this is a requirement of fused types of cython
        centers = numpy.array(init, dtype=X.dtype)
    elif callable(init):
        centers = init(norm, X, k, random_state=random_state)
        centers = numpy.asarray(centers, dtype=X.dtype)
    else:
        raise ValueError("the init parameter for the k-means should "
                         "be 'k-means++' or 'random' or an ndarray, "
                         "'%s' (type '%s') was passed." % (init, type(init)))

    if issparse(centers):
        centers = centers.toarray()

    def _validate_center_shape(X, k, centers):
        """Check if centers is compatible with X and n_clusters"""
        if centers.shape[0] != k:
            raise ValueError(
                f"The shape of the initial centers {centers.shape} does not "
                f"match the number of clusters {k}.")
        if centers.shape[1] != X.shape[1]:
            raise ValueError(
                f"The shape of the initial centers {centers.shape} does not "
                f"match the number of features of the data {X.shape[1]}.")

    _validate_center_shape(X, k, centers)
    return centers


def _centers_dense(X, sample_weight, labels, n_clusters, distances,
                   X_sort_index):
    """
    M step of the K-means EM algorithm.
    Computation of cluster centers / means.

    :param X: array-like, shape (n_samples, n_features)
    :param sample_weight: array-like, shape (n_samples,)
        The weights for each observation in X.
    :param labels: array of integers, shape (n_samples)
        Current label assignment
    :param n_clusters: int
        Number of desired clusters
    :param distances: array-like, shape (n_samples)
        Distance to closest cluster for each sample.
    :param X_sort_index: array-like, shape (n_samples, n_features)
        index of each feature in all features
    :return: centers, array, shape (n_clusters, n_features)
        The resulting centers
    """
    dtype = X.dtype
    n_features = X.shape[1]
    n_samples = X.shape[0]

    centers = numpy.zeros((n_clusters, n_features), dtype=dtype)
    weight_in_cluster = numpy.zeros((n_clusters,), dtype=dtype)

    for i in range(n_samples):
        c = labels[i]
        weight_in_cluster[c] += sample_weight[i]
    empty_clusters = numpy.where(weight_in_cluster == 0)[0]

    if len(empty_clusters) > 0:  # pragma: no cover
        # find points to reassign empty clusters to
        far_from_centers = distances.argsort()[::-1]

        for i, cluster_id in enumerate(empty_clusters):
            far_index = far_from_centers[i]
            new_center = X[far_index] * sample_weight[far_index]
            centers[cluster_id] = new_center
            weight_in_cluster[cluster_id] = sample_weight[far_index]

    if sample_weight.min() == sample_weight.max():
        # to optimize
        for i in range(n_clusters):
            sub = X[labels == i]
            med = numpy.median(sub, axis=0)
            centers[i, :] = med
    else:
        raise NotImplementedError(  # pragma: no cover
            "Non uniform weights are not implemented yet as "
            "the cost would be very high. "
            "See https://en.wikipedia.org/wiki/Weighted_median#Algorithm.")
    return centers


def _kmeans_single_lloyd(norm, X, sample_weight, n_clusters, max_iter=300,
                         init='k-means++', verbose=False,
                         random_state=None, tol=1e-4,
                         precompute_distances=True):
    """
    A single run of k-means, assumes preparation completed prior.

    :param norm: 'l1' or 'l2'
    :param X: array-like of floats, shape (n_samples, n_features)
        The observations to cluster.
    :param n_clusters: int
        The number of clusters to form as well as the number of
        centroids to generate.
    :param sample_weight: array-like, shape (n_samples,)
        The weights for each observation in X.
    :param max_iter: int, optional, default 300
        Maximum number of iterations of the k-means algorithm to run.
    :param init: {'k-means++', 'random', or ndarray, or a callable}, optional
        Method for initialization, default to 'k-means++':

        'k-means++' : selects initial cluster centers for k-mean
        clustering in a smart way to speed up convergence. See section
        Notes in k_init for more details.

        'random': choose k observations (rows) at random from data for
        the initial centroids.

        If an ndarray is passed, it should be of shape (k, p) and gives
        the initial centers.

        If a callable is passed, it should take arguments X, k and
        and a random state and return an initialization.

    :param tol: float, optional
        The relative increment in the results before declaring convergence.
    :param verbose: boolean, optional
        Verbosity mode
    :param precompute_distances: boolean, default: True
        Precompute distances (faster but takes more memory).
    :param random_state: int, RandomState instance or None (default)
        Determines random number generation for centroid initialization. Use
        an int to make the randomness deterministic.
        See :term:`Glossary <random_state>`.
    :return: centroid : float ndarray with shape (k, n_features)
        Centroids found at the last iteration of k-means.
    :return: label : integer ndarray with shape (n_samples,)
        label[i] is the code or index of the centroid the
        i'th observation is closest to.
    :return: inertia : float
        The final value of the inertia criterion (sum of squared distances to
        the closest centroid for all observations in the training set).
    :return: n_iter : int
        Number of iterations run.
    """
    random_state = check_random_state(random_state)

    sample_weight = _check_sample_weight(sample_weight, X)

    best_labels, best_inertia, best_centers = None, None, None
    # init
    centers = _init_centroids(
        norm, X, n_clusters, init, random_state=random_state)
    if verbose:  # pragma no cover
        print("Initialization complete")

    # Allocate memory to store the distances for each sample to its
    # closer center for reallocation in case of ties
    distances = numpy.zeros(shape=(X.shape[0],), dtype=X.dtype)
    X_sort_index = numpy.argsort(X, axis=0)

    # iterations
    for i in range(max_iter):
        centers_old = centers.copy()
        # labels assignment is also called the E-step of EM
        labels, inertia = \
            _labels_inertia(norm, X, sample_weight, centers,
                            precompute_distances=precompute_distances,
                            distances=distances)

        # computation of the means is also called the M-step of EM
        centers = _centers_dense(X, sample_weight, labels, n_clusters, distances,
                                 X_sort_index)

        if verbose:  # pragma no cover
            print("Iteration %2d, inertia %.3f" % (i, inertia))

        if best_inertia is None or inertia < best_inertia:
            best_labels = labels.copy()
            best_centers = centers.copy()
            best_inertia = inertia

        center_shift_total = numpy.sum(
            numpy.abs(centers_old - centers).ravel())
        if center_shift_total <= tol:
            if verbose:  # pragma no cover
                print("Converged at iteration %d: "
                      "center shift %r within tolerance %r"
                      % (i, center_shift_total, tol))
            break

    if center_shift_total > 0:
        # rerun E-step in case of non-convergence so that predicted labels
        # match cluster centers
        best_labels, best_inertia = \
            _labels_inertia(norm, X, sample_weight, best_centers,
                            precompute_distances=precompute_distances,
                            distances=distances)

    return best_labels, best_inertia, best_centers, i + 1


def _labels_inertia(norm, X, sample_weight, centers,
                    precompute_distances=True, distances=None):
    """
    E step of the K-means EM algorithm.

    Computes the labels and the inertia of the given samples and centers.
    This will compute the distances in-place.

    :param norm: 'l1' or 'l2'
    :param X: float64 array-like or CSR sparse matrix, shape (n_samples, n_features)
        The input samples to assign to the labels.
    :param sample_weight: array-like, shape (n_samples,)
        The weights for each observation in X.
    :param centers: float array, shape (k, n_features)
        The cluster centers.
    :param precompute_distances: boolean, default: True
        Precompute distances (faster but takes more memory).
    :param distances: existing distances
    :return: labels : int array of shape(n)
        The resulting assignment
    :return: inertia : float
        Sum of squared distances of samples to their closest cluster center.
    """
    if norm == 'l2':
        return _labels_inertia_skl(
            X, sample_weight=sample_weight, centers=centers,
            precompute_distances=precompute_distances,
            x_squared_norms=None)

    sample_weight = _check_sample_weight(sample_weight, X)
    # set the default value of centers to -1 to be able to detect any anomaly
    # easily
    if distances is None:
        distances = numpy.zeros(shape=(0,), dtype=X.dtype)
    # distances will be changed in-place
    if issparse(X):
        raise NotImplementedError(  # pragma no cover
            "Sparse matrix is not implemented for norm 'l1'.")
    if precompute_distances:
        return _labels_inertia_precompute_dense(
            norm=norm, X=X, sample_weight=sample_weight,
            centers=centers, distances=distances)
    raise NotImplementedError(  # pragma no cover
        "precompute_distances is False, not implemented for norm 'l1'.")


def _tolerance(norm, X, tol):
    """Return a tolerance which is independent of the dataset"""
    if norm == 'l2':
        return _tolerance_skl(X, tol)
    if norm == 'l1':
        variances = numpy.sum(numpy.abs(X), axis=0) / X.shape[0]
        return variances.sum()
    raise NotImplementedError(  # pragma no cover
        "not implemented for norm '{}'.".format(norm))


class KMeansL1L2(KMeans):
    """
    K-Means clustering with either norm L1 or L2.
    See notebook :ref:`kmeansl1rst` for an example.

    :param n_clusters: int, default=8
        The number of clusters to form as well as the number of
        centroids to generate.
    :param init: {'k-means++', 'random'} or ndarray of shape \
            (n_clusters, n_features), default='k-means++'
        Method for initialization, defaults to 'k-means++':

        'k-means++' : selects initial cluster centers for k-mean
        clustering in a smart way to speed up convergence. See section
        Notes in k_init for more details.

        'random': choose k observations (rows) at random from data for
        the initial centroids.

        If an ndarray is passed, it should be of shape (n_clusters, n_features)
        and gives the initial centers.

    :param n_init: int, default=10
        Number of time the k-means algorithm will be run with different
        centroid seeds. The final results will be the best output of
        n_init consecutive runs in terms of inertia.
    :param max_iter: int, default=300
        Maximum number of iterations of the k-means algorithm for a
        single run.
    :param tol: float, default=1e-4
        Relative tolerance with regards to inertia to declare convergence.
    :param precompute_distances: 'auto' or bool, default='auto'
        Precompute distances (faster but takes more memory).

        'auto' : do not precompute distances if n_samples * n_clusters > 12
        million. This corresponds to about 100MB overhead per job using
        double precision.

        True : always precompute distances.

        False : never precompute distances.

    :param verbose: int, default=0
        Verbosity mode.
    :param random_state: int, RandomState instance, default=None
        Determines random number generation for centroid initialization. Use
        an int to make the randomness deterministic.
        See :term:`Glossary <random_state>`.
    :param copy_x: bool, default=True
        When pre-computing distances it is more numerically accurate to center
        the data first.  If copy_x is True (default), then the original data is
        not modified, ensuring X is C-contiguous.  If False, the original data
        is modified, and put back before the function returns, but small
        numerical differences may be introduced by subtracting and then adding
        the data mean, in this case it will also not ensure that data is
        C-contiguous which may cause a significant slowdown.
    :param n_jobs: int, default=None
        The number of jobs to use for the computation. This works by computing
        each of the n_init runs in parallel.

        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.
    :param algorithm: {"auto", "full", "elkan"}, default="auto"
        K-means algorithm to use. The classical EM-style algorithm is "full".
        The "elkan" variation is more efficient by using the triangle
        inequality, but currently doesn't support sparse data. "auto" chooses
        "elkan" for dense data and "full" for sparse data.
    :param norm: {"L1", "L2"}
        The norm *L2* is identical to :epkg:`KMeans`.
        Norm *L1* uses a complete different path.

    Fitted attributes:

    * `cluster_centers_`: ndarray of shape (n_clusters, n_features)
        Coordinates of cluster centers. If the algorithm stops before fully
        converging (see ``tol`` and ``max_iter``), these will not be
        consistent with ``labels_``.
    * `labels_`: ndarray of shape (n_samples,)
        Labels of each point
    * `inertia_`: float
        Sum of squared distances of samples to their closest cluster center.
    * `n_iter_`: int
        Number of iterations run.
    """

    def __init__(self, n_clusters=8, init='k-means++', n_init=10,
                 max_iter=300, tol=1e-4, precompute_distances='auto',
                 verbose=0, random_state=None, copy_x=True,
                 n_jobs=None, algorithm='full', norm='L2'):

        KMeans.__init__(self, n_clusters=n_clusters, init=init, n_init=n_init,
                        max_iter=max_iter, tol=tol,
                        precompute_distances=precompute_distances,
                        verbose=verbose, random_state=random_state,
                        copy_x=copy_x, n_jobs=n_jobs, algorithm=algorithm)
        self.norm = norm.lower()
        if self.norm == 'l1' and self.algorithm != 'full':
            raise NotImplementedError(  # pragma no cover
                "Only algorithm 'full' is implemented with norm 'l1'.")

    def fit(self, X, y=None, sample_weight=None):
        """
        Computes k-means clustering.

        :param X: array-like or sparse matrix, shape=(n_samples, n_features)
            Training instances to cluster. It must be noted that the data
            will be converted to C ordering, which will cause a memory
            copy if the given data is not C-contiguous.
        :param y: Ignored
            Not used, present here for API consistency by convention.
        :param sample_weight: array-like, shape (n_samples,), optional
            The weights for each observation in X. If None, all observations
            are assigned equal weight (default: None).
        :return: self
            Fitted estimator.
        """
        if self.norm == 'l2':
            KMeans.fit(self, X=X, y=y, sample_weight=sample_weight)
        elif self.norm == 'l1':
            self._fit_l1(X=X, y=y, sample_weight=sample_weight)
        else:
            raise NotImplementedError(  # pragma no cover
                "Norm is not L1 or L2 but '{}'.".format(self.norm))
        return self

    def _fit_l1(self, X, y=None, sample_weight=None):
        """
        Computes k-means clustering with norm `'l1'`.

        :param X: array-like or sparse matrix, shape=(n_samples, n_features)
            Training instances to cluster. It must be noted that the data
            will be converted to C ordering, which will cause a memory
            copy if the given data is not C-contiguous.
        :param y: Ignored
            Not used, present here for API consistency by convention.
        :param sample_weight: array-like, shape (n_samples,), optional
            The weights for each observation in X. If None, all observations
            are assigned equal weight (default: None).
        :return: self
            Fitted estimator.
        """
        random_state = check_random_state(self.random_state)

        n_init = self.n_init
        if n_init <= 0:
            raise ValueError(  # pragma no cover
                "Invalid number of initializations."
                " n_init=%d must be bigger than zero." % n_init)

        if self.max_iter <= 0:
            raise ValueError(  # pragma no cover
                'Number of iterations should be a positive number,'
                ' got %d instead' % self.max_iter)

        # avoid forcing order when copy_x=False
        order = "C" if self.copy_x else None
        X = check_array(X, accept_sparse='csr', dtype=[numpy.float64, numpy.float32],
                        order=order, copy=self.copy_x)
        # verify that the number of samples given is larger than k
        if _num_samples(X) < self.n_clusters:
            raise ValueError(  # pragma no cover
                "n_samples=%d should be >= n_clusters=%d" % (
                    _num_samples(X), self.n_clusters))

        tol = _tolerance(self.norm, X, self.tol)

        # If the distances are precomputed every job will create a matrix of
        # shape (n_clusters, n_samples). To stop KMeans from eating up memory
        # we only activate this if the created matrix is guaranteed to be
        # under 100MB. 12 million entries consume a little under 100MB if they
        # are of type double.
        precompute_distances = self.precompute_distances
        if precompute_distances == 'auto':
            n_samples = X.shape[0]
            precompute_distances = (self.n_clusters * n_samples) < 12e6
        elif isinstance(precompute_distances, bool):  # pragma: no cover
            pass
        else:
            raise ValueError(  # pragma no cover
                "precompute_distances should be 'auto' or True/False"
                ", but a value of %r was passed" % precompute_distances)

        # Validate init array
        init = self.init
        if hasattr(init, '__array__'):
            init = check_array(init, dtype=X.dtype.type, copy=True)
            if hasattr(self, '_validate_center_shape'):
                self._validate_center_shape(  # pylint: disable=E1101
                    X, init)

            if n_init != 1:
                warnings.warn(
                    'Explicit initial center position passed: '
                    'performing only one init in k-means instead of n_init=%d'
                    % n_init, RuntimeWarning, stacklevel=2)
                n_init = 1

        best_labels, best_inertia, best_centers = None, None, None
        algorithm = self.algorithm
        if self.n_clusters == 1:
            # elkan doesn't make sense for a single cluster, full will produce
            # the right result.
            algorithm = "full"  # pragma: no cover
        if algorithm == "auto":
            algorithm = "full"  # pragma: no cover
        if algorithm == "full":
            kmeans_single = _kmeans_single_lloyd
        else:
            raise ValueError(  # pragma no cover
                "Algorithm must be 'auto', 'full' or 'elkan', got"
                " %s" % str(algorithm))

        seeds = random_state.randint(numpy.iinfo(numpy.int32).max, size=n_init)
        if effective_n_jobs(self.n_jobs) == 1:
            # For a single thread, less memory is needed if we just store one
            # set of the best results (as opposed to one set per run per
            # thread).
            for seed in seeds:
                # run a k-means once
                labels, inertia, centers, n_iter_ = kmeans_single(
                    self.norm, X, sample_weight, self.n_clusters,
                    max_iter=self.max_iter, init=init, verbose=self.verbose,
                    precompute_distances=precompute_distances, tol=tol,
                    random_state=seed)
                # determine if these results are the best so far
                if best_inertia is None or inertia < best_inertia:
                    best_labels = labels.copy()
                    best_centers = centers.copy()
                    best_inertia = inertia
                    best_n_iter = n_iter_
        else:
            # parallelisation of k-means runs
            results = Parallel(n_jobs=self.n_jobs, verbose=0)(
                delayed(kmeans_single)(
                    self.norm, X, sample_weight, self.n_clusters,
                    max_iter=self.max_iter, init=init,
                    verbose=self.verbose, tol=tol,
                    precompute_distances=precompute_distances,
                    # Change seed to ensure variety
                    random_state=seed
                )
                for seed in seeds)
            # Get results with the lowest inertia
            labels, inertia, centers, n_iters = zip(*results)
            best = numpy.argmin(inertia)
            best_labels = labels[best]
            best_inertia = inertia[best]
            best_centers = centers[best]
            best_n_iter = n_iters[best]

        distinct_clusters = len(set(best_labels))
        if distinct_clusters < self.n_clusters:
            warnings.warn(  # pragma no cover
                "Number of distinct clusters ({}) found smaller than "
                "n_clusters ({}). Possibly due to duplicate points "
                "in X.".format(distinct_clusters, self.n_clusters),
                ConvergenceWarning, stacklevel=2)

        self.cluster_centers_ = best_centers
        self.labels_ = best_labels
        self.inertia_ = best_inertia
        self.n_iter_ = best_n_iter
        return self

    def transform(self, X):
        """
        Transforms *X* to a cluster-distance space.

        In the new space, each dimension is the distance to the cluster
        centers.  Note that even if X is sparse, the array returned by
        `transform` will typically be dense.

        :param X: {array-like, sparse matrix} of shape (n_samples, n_features)
            New data to transform.
        :return: X_new : array, shape [n_samples, k]
            X transformed in the new space.
        """
        if self.norm == 'l2':
            return KMeans.transform(self, X)
        if self.norm == 'l1':
            return self._transform_l1(X)
        raise NotImplementedError(  # pragma no cover
            "Norm is not L1 or L2 but '{}'.".format(self.norm))

    def _transform_l1(self, X):
        """
        Returns the distance of each point in *X* to
        every fit clusters.
        """
        check_is_fitted(self)
        X = self._check_test_data(X)
        return manhattan_distances(X, self.cluster_centers_)

    def predict(self, X, sample_weight=None):
        """
        Predicts the closest cluster each sample in X belongs to.

        In the vector quantization literature, `cluster_centers_` is called
        the code book and each value returned by `predict` is the index of
        the closest code in the code book.

        :param X: {array-like, sparse matrix} of shape (n_samples, n_features)
            New data to predict.
        :param sample_weight: array-like, shape (n_samples,), optional
            The weights for each observation in X. If None, all observations
            are assigned equal weight (default: None), unused here
        :return: labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """
        if self.norm == 'l2':
            return KMeans.predict(self, X)
        if self.norm == 'l1':
            return self._predict_l1(X, sample_weight=sample_weight)
        raise NotImplementedError(  # pragma no cover
            "Norm is not L1 or L2 but '{}'.".format(self.norm))

    def _predict_l1(self, X, sample_weight=None, return_distances=False):
        """
        Returns the distance of each point in *X* to
        every fit clusters.

        :param X: features
        :param sample_weight: (unused)
        :param return_distances: returns distances as well
        :return: labels or `labels, distances`
        """
        labels, mindist = pairwise_distances_argmin_min(
            X=X, Y=self.cluster_centers_, metric='manhattan')
        labels = labels.astype(numpy.int32, copy=False)
        if return_distances:
            return labels, mindist
        return labels



"""
@file
@brief Shortcuts to *search_rank*.
"""

from .search_engine_predictions import SearchEnginePredictions
from .search_engine_predictions_images import SearchEnginePredictionImages
from .search_engine_vectors import SearchEngineVectors



# pylint: disable=C0302
"""
@file
@brief Implements k-means with norms L1 and L2.
"""
import numpy
from scipy.sparse import issparse
# Source: https://github.com/scikit-learn/scikit-learn/blob/95d4f0841d57e8b5f6b2a570312e9d832e69debc/sklearn/cluster/_k_means_fast.pyx
from sklearn.utils.sparsefuncs_fast import assign_rows_csr  # pylint: disable=W0611,E0611
try:
    from sklearn.cluster._kmeans import _check_sample_weight
except ImportError:
    from sklearn.cluster._kmeans import (
        _check_normalize_sample_weight as _check_sample_weight)
from sklearn.metrics.pairwise import pairwise_distances_argmin_min


def _labels_inertia_precompute_dense(norm, X, sample_weight, centers, distances):
    """
    Computes labels and inertia using a full distance matrix.

    This will overwrite the 'distances' array in-place.

    :param norm: 'l1' or 'l2'
    :param X:  numpy array, shape (n_sample, n_features)
        Input data.
    :param sample_weight: array-like, shape (n_samples,)
        The weights for each observation in X.
    :param centers: numpy array, shape (n_clusters, n_features)
        Cluster centers which data is assigned to.
    :param distances: numpy array, shape (n_samples,)
        Pre-allocated array in which distances are stored.
    :return: labels : numpy array, dtype=numpy.int, shape (n_samples,)
        Indices of clusters that samples are assigned to.
    :return: inertia : float
        Sum of squared distances of samples to their closest
        cluster center.
    """
    n_samples = X.shape[0]
    if norm == 'l2':
        labels, mindist = pairwise_distances_argmin_min(
            X=X, Y=centers, metric='euclidean', metric_kwargs={'squared': True})
    elif norm == 'l1':
        labels, mindist = pairwise_distances_argmin_min(
            X=X, Y=centers, metric='manhattan')
    else:  # pragma no cover
        raise NotImplementedError(
            "Not implemented for norm '{}'.".format(norm))
    # cython k-means code assumes int32 inputs
    labels = labels.astype(numpy.int32, copy=False)
    if n_samples == distances.shape[0]:
        # distances will be changed in-place
        distances[:] = mindist
    inertia = (mindist * sample_weight).sum()
    return labels, inertia


def _assign_labels_csr(X, sample_weight, x_squared_norms, centers,
                       labels, distances):
    """Compute label assignment and inertia for a CSR input
    Return the inertia (sum of squared distances to the centers).
    """
    if (distances is not None and
            distances.shape != (X.shape[0], )):
        raise ValueError(  # pragma: no cover
            "Dimension mismatch for distance got {}, expecting {}."
            "".format(distances.shape, (X.shape[0], centers.shape[0])))
    n_clusters = centers.shape[0]
    n_samples = X.shape[0]
    store_distances = 0
    inertia = 0.0

    if centers.dtype == numpy.float32:
        center_squared_norms = numpy.zeros(n_clusters, dtype=numpy.float32)
    else:
        center_squared_norms = numpy.zeros(n_clusters, dtype=numpy.float64)

    if n_samples == distances.shape[0]:
        store_distances = 1

    for center_idx in range(n_clusters):
        center_squared_norms[center_idx] = numpy.dot(
            centers[center_idx, :], centers[center_idx, :])

    for sample_idx in range(n_samples):
        min_dist = -1
        for center_idx in range(n_clusters):
            dist = 0.0
            # hardcoded: minimize euclidean distance to cluster center:
            # ||a - b||^2 = ||a||^2 + ||b||^2 -2 <a, b>
            dist += X[sample_idx, :] @ centers[center_idx, :].reshape((-1, 1))
            dist *= -2
            dist += center_squared_norms[center_idx]
            dist += x_squared_norms[sample_idx]
            dist *= sample_weight[sample_idx]
            if min_dist == -1 or dist < min_dist:
                min_dist = dist
                labels[sample_idx] = center_idx
                if store_distances:
                    distances[sample_idx] = dist
        inertia += min_dist

    return inertia


def _assign_labels_array(X, sample_weight, x_squared_norms, centers,
                         labels, distances):
    """Compute label assignment and inertia for a dense array
    Return the inertia (sum of squared distances to the centers).
    """
    n_clusters = centers.shape[0]
    n_samples = X.shape[0]
    store_distances = 0
    inertia = 0.0

    if centers.dtype == numpy.float32:
        center_squared_norms = numpy.zeros(n_clusters, dtype=numpy.float32)
    else:
        center_squared_norms = numpy.zeros(n_clusters, dtype=numpy.float64)

    if n_samples == distances.shape[0]:
        store_distances = 1

    for center_idx in range(n_clusters):
        center_squared_norms[center_idx] = numpy.dot(
            centers[center_idx, :], centers[center_idx, :])

    for sample_idx in range(n_samples):
        min_dist = -1
        for center_idx in range(n_clusters):
            dist = 0.0
            # hardcoded: minimize euclidean distance to cluster center:
            # ||a - b||^2 = ||a||^2 + ||b||^2 -2 <a, b>
            dist += numpy.dot(X[sample_idx, :], centers[center_idx, :])
            dist *= -2
            dist += center_squared_norms[center_idx]
            dist += x_squared_norms[sample_idx]
            dist *= sample_weight[sample_idx]
            if min_dist == -1 or dist < min_dist:
                min_dist = dist
                labels[sample_idx] = center_idx

        if store_distances:
            distances[sample_idx] = min_dist
        inertia += min_dist

    return inertia


def _labels_inertia_skl(X, sample_weight, x_squared_norms, centers,
                        precompute_distances=True, distances=None):
    """E step of the K-means EM algorithm.
    Compute the labels and the inertia of the given samples and centers.
    This will compute the distances in-place.

    :param X: float64 array-like or CSR sparse matrix, shape (n_samples, n_features)
        The input samples to assign to the labels.
    :param sample_weight: array-like, shape (n_samples,)
        The weights for each observation in X.
    :param x_squared_norms: array, shape (n_samples,)
        Precomputed squared euclidean norm of each data point, to speed up
        computations.
    :param centers: float array, shape (k, n_features)
        The cluster centers.
    :param precompute_distances: boolean, default: True
        Precompute distances (faster but takes more memory).
    :param distances: float array, shape (n_samples,)
        Pre-allocated array to be filled in with each sample's distance
        to the closest center.
    :return: labels, int array of shape(n)
        The resulting assignment
    :return: inertia, float
        Sum of squared distances of samples to their closest cluster center.
    """
    n_samples = X.shape[0]
    sample_weight = _check_sample_weight(sample_weight, X)
    # set the default value of centers to -1 to be able to detect any anomaly
    # easily
    labels = numpy.full(n_samples, -1, numpy.int32)
    if distances is None:
        distances = numpy.zeros(shape=(0,), dtype=X.dtype)
    # distances will be changed in-place
    if issparse(X):
        inertia = _assign_labels_csr(
            X, sample_weight, x_squared_norms, centers, labels,
            distances=distances)
    else:
        if precompute_distances:
            return _labels_inertia_precompute_dense(
                norm='l2', X=X, sample_weight=sample_weight,
                centers=centers, distances=distances)
        inertia = _assign_labels_array(
            X, sample_weight, x_squared_norms, centers, labels,
            distances=distances)
    return labels, inertia


def _centers_dense(X, sample_weight, labels, n_clusters, distances):
    """
    M step of the K-means EM algorithm
    Computation of cluster centers / means.

    :param X: array-like, shape (n_samples, n_features)
    :param sample_weight: array-like, shape (n_samples,)
        The weights for each observation in X.
    :param labels: array of integers, shape (n_samples)
        Current label assignment
    :param n_clusters: int
        Number of desired clusters
    :param distances: array-like, shape (n_samples)
        Distance to closest cluster for each sample.
    :return: centers : array, shape (n_clusters, n_features)
        The resulting centers
    """
    n_samples = X.shape[0]
    n_features = X.shape[1]

    dtype = X.dtype
    centers = numpy.zeros((n_clusters, n_features), dtype=dtype)
    weight_in_cluster = numpy.zeros((n_clusters,), dtype=dtype)

    for i in range(n_samples):
        c = labels[i]
        weight_in_cluster[c] += sample_weight[i]
    empty_clusters = numpy.where(weight_in_cluster == 0)[0]
    # maybe also relocate small clusters?

    if distances is not None and len(empty_clusters):
        # find points to reassign empty clusters to
        far_from_centers = distances.argsort()[::-1]

        for i, cluster_id in enumerate(empty_clusters):
            far_index = far_from_centers[i]
            new_center = X[far_index] * sample_weight[far_index]
            centers[cluster_id] = new_center
            weight_in_cluster[cluster_id] = sample_weight[far_index]

    for i in range(n_samples):
        for j in range(n_features):
            centers[labels[i], j] += X[i, j] * sample_weight[i]

    centers /= weight_in_cluster[:, numpy.newaxis]

    return centers


def _centers_sparse(X, sample_weight, labels, n_clusters, distances):
    """
    M step of the K-means EM algorithm
    Computation of cluster centers / means.

    :param X: scipy.sparse.csr_matrix, shape (n_samples, n_features)
    :param sample_weight: array-like, shape (n_samples,)
        The weights for each observation in X.
    :param labels: array of integers, shape (n_samples)
        Current label assignment
    :param n_clusters: int
        Number of desired clusters
    :param distances: array-like, shape (n_samples)
        Distance to closest cluster for each sample.
    :return: centers, array, shape (n_clusters, n_features)
        The resulting centers
    """
    n_samples = X.shape[0]
    n_features = X.shape[1]

    data = X.data
    indices = X.indices
    indptr = X.indptr

    dtype = X.dtype
    centers = numpy.zeros((n_clusters, n_features), dtype=dtype)
    weight_in_cluster = numpy.zeros((n_clusters,), dtype=dtype)
    for i in range(n_samples):
        c = labels[i]
        weight_in_cluster[c] += sample_weight[i]
    empty_clusters = numpy.where(weight_in_cluster == 0)[0]
    n_empty_clusters = empty_clusters.shape[0]

    # maybe also relocate small clusters?

    if n_empty_clusters > 0:
        # find points to reassign empty clusters to
        far_from_centers = distances.argsort()[::-1][:n_empty_clusters]
        assign_rows_csr(X, far_from_centers, empty_clusters, centers)

        for i in range(n_empty_clusters):
            weight_in_cluster[empty_clusters[i]] = 1

    for i in range(labels.shape[0]):
        curr_label = labels[i]
        for ind in range(indptr[i], indptr[i + 1]):
            j = indices[ind]
            centers[curr_label, j] += data[ind] * sample_weight[i]

    centers /= weight_in_cluster[:, numpy.newaxis]

    return centers



# -*- coding: utf-8 -*-
"""
@file
@brief Implements a *learner* or a *transform* which follows the same API
as every :epkg:`scikit-learn` transform.
"""
import textwrap
import warnings
from .sklearn_parameters import SkLearnParameters


class SkBase:
    """
    Pattern of a *learner* or a *transform* which follows the API
    of :epkg:`scikit-learn`.
    """

    def __init__(self, **kwargs):
        """
        Stores the parameters, see
        @see cl SkLearnParameters, it keeps a copy of
        the parameters to easily implements method *get_params*
        and clones a model.
        """
        self.P = SkLearnParameters(**kwargs)

    def fit(self, X, y=None, sample_weight=None):
        """
        Trains a model.

        @param      X               features
        @param      y               target
        @param      sample_weight   weight
        @return                     self
        """
        raise NotImplementedError()  # pragma: no cover

    def get_params(self, deep=True):
        """
        Returns the parameters which define the objet,
        all are needed to clone the object.

        @param      deep        unused here
        @return                 dict
        """
        return self.P.to_dict()

    def set_params(self, **values):
        """
        Udpates parameters which define the object,
        all needed to clone the object.

        @param      values      values
        @return                 dictionary
        """
        self.P = SkLearnParameters(**values)
        return self

    def __eq__(self, o):
        """
        Compares two objects, more precisely,
        compares the parameters which define the object.
        """
        return self.test_equality(o, False)

    def test_equality(self, o, exc=True):
        """
        Compares two objects and checks parameters have
        the same values.

        @param      p1      dictionary
        @param      p2      dictionary
        @param      exc     raises an exception if there is a difference
        @return             boolean
        """
        if self.__class__ != o.__class__:
            return False
        p1 = self.get_params()
        p2 = o.get_params()
        return SkBase.compare_params(p1, p2, exc=exc)

    @staticmethod
    def compare_params(p1, p2, exc=True):
        """
        Compares two sets of parameters.

        @param      p1      dictionary
        @param      p2      dictionary
        @param      exc     raises an exception if error is met
        @return             boolean
        """
        if p1 == p2:
            return True
        for k in p1:
            if k not in p2:
                if exc:
                    raise KeyError("Key '{0}' was removed.".format(k))
                else:
                    return False
        for k in p2:
            if k not in p1:
                if exc:
                    raise KeyError("Key '{0}' was added.".format(k))
                return False
        for k in sorted(p1):
            v1, v2 = p1[k], p2[k]
            if hasattr(v1, 'test_equality'):
                b = v1.test_equality(v2, exc=exc)
                if exc and v1 is not v2:
                    warnings.warn(  # pragma: no cover
                        "v2 is a clone of v1 not v1 itself for key '{0}' and class {1}."
                        "".format(k, type(v1)))
            elif isinstance(v1, list) and isinstance(v2, list) and len(v1) == len(v2):
                b = True
                for e1, e2 in zip(v1, v2):
                    if hasattr(e1, 'test_equality'):
                        b = e1.test_equality(e2, exc=exc)
                        if not b:
                            return b
            elif isinstance(v1, dict) and isinstance(v2, dict) and set(v1) == set(v2):
                b = True
                for e1, e2 in zip(sorted(v1.items()), sorted(v2.items())):
                    if hasattr(e1[1], 'test_equality'):
                        b = e1[1].test_equality(e2[1], exc=exc)
                        if not b:
                            return b
                    elif e1[1] != e2[1]:
                        return False
            elif hasattr(v1, "get_params") and hasattr(v2, "get_params"):
                b = SkBase.compare_params(v1.get_params(
                    deep=False), v2.get_params(deep=False), exc=exc)
            else:
                b = v1 == v2
            if not b:
                if exc:
                    raise ValueError(
                        "Values for key '{0}' are different.\n---\n{1}\n---\n{2}".format(k, v1, v2))
                else:
                    return False
        return True

    def __repr__(self):
        """
        usual
        """
        res = "{0}({1})".format(self.__class__.__name__, str(self.P))
        return "\n".join(textwrap.wrap(res, subsequent_indent="    "))



"""
@file
@brief Find patterns in timeseries.
"""
import numpy
import pandas
from sklearn.cluster import KMeans
from .agg import aggregate_timeseries


def find_ts_group_pattern(ttime, values, names, name_subset=None,
                          per='week', unit='half-hour', agg='sum',
                          estimator=None, fLOG=None):
    """
    Clusters times series to find similar patterns.

    @param      ttime           time column
    @param      values          features to use to cluster
    @param      names           column which holds group name
    @param      name_subset     subset of groups to study, None for all
    @param      per             aggragation per week
    @param      estimator       estimator used to find pattern,
                                :epkg:`sklearn:cluster:KMeans` and
                                10 groups
    @param      fLOG            logging function
    @return                     found clusters, distances
    """
    for var, na in zip([ttime, values, names], ['ttime', 'values', 'names']):
        if not isinstance(var, numpy.ndarray):
            raise TypeError("'{}' must an array not {}".format(na, type(var)))
    # builds features
    set_names = set(names)
    if name_subset is not None:
        set_names &= set(name_subset)
    if fLOG:
        fLOG(  # pragma: no cover
            '[find_ts_group_pattern] build features, {} groups'.format(len(set_names)))
    gr_names = []
    to_merge = []
    for name in set_names:
        indices = names == name
        gr_ttime = ttime[indices]
        gr_values = values[indices]
        gr = aggregate_timeseries(None, gr_ttime, gr_values,
                                  unit=unit, agg=agg, per=per)
        gr.set_index(gr.columns[0], inplace=True)
        gr_names.append(name)
        to_merge.append(gr)

    if fLOG:
        fLOG(  # pragma: no cover
            '[find_ts_group_pattern] merge features')
    all_merged = pandas.concat(to_merge, axis=1)
    all_merged.fillna(0, inplace=True)
    ncol = all_merged.shape[1] // len(gr_names)
    gr_feats = []
    for i, name in enumerate(gr_names):
        feats = all_merged.iloc[:, i * ncol: (i + 1) * ncol].values.ravel()
        gr_feats.append(feats)

    gr_feats = numpy.vstack(gr_feats)

    # cluster
    if fLOG:
        fLOG(  # pragma: no cover
            '[find_ts_group_pattern] clustering, shape={}'.format(gr_feats.shape))
    if estimator is None:
        estimator = KMeans()
    estimator.fit(gr_feats)

    # predicted clusters
    pred = estimator.predict(gr_feats)
    dist = estimator.transform(gr_feats)
    if fLOG:
        fLOG(  # pragma: no cover
            '[find_ts_group_pattern] number of clusters: {}'.format(len(set(pred))))

    row_name = {n: i for i, n in enumerate(gr_names)}
    clusters = numpy.empty(ttime.shape[0], dtype=pred.dtype)
    dists = numpy.empty((ttime.shape[0], dist.shape[1]), dtype=dist.dtype)

    for i in range(ttime.shape[0]):
        if names[i] in row_name:
            index = row_name[names[i]]
            clusters[i] = pred[index]
            dists[i, :] = dist[index, :]
        else:
            clusters[i] = -1
            dists[i, :] = numpy.nan

    return clusters, dists
